# -*- coding: latin-1 -*-
from morphonet.plugins import MorphoPlugin
from .functions import *

class SplitOnRaw(MorphoPlugin):
    """ This plugin split opbjects based local maximum found at a miniym distance on the rawdata
    For more information see https://scikit-image.org/docs/dev/auto_examples/segmentation/plot_peak_local_max.html

    Parameters
    ----------
    Objects: 
        It can be apply either on selected or colored objects
    Volume Minimum : int , default 2000
        The minimum accepted volume to create a new object
    Min Distance: int, default 20
        The minmum distance between two peaks in the peak_local_max fucntion
    """

    def __init__(self): #PLUGIN DEFINITION 
        MorphoPlugin.__init__(self) 
        self.set_name("Split On Raw")
        self.add_inputfield("Volume Minimum",default=2000)
        self.add_inputfield("Min Distance",default=20)
        self.set_parent("Split objects")

    def process(self,t,dataset,objects): #PLUGIN EXECUTION
        if not self.start(t,dataset,objects):
            return None
            
        from skimage.segmentation import watershed
        from skimage.feature import peak_local_max
        import numpy as np
        min_distance=int(self.get_inputfield("Min Distance"))
        minVol=int(self.get_inputfield("Volume Minimum"))
        for cid in objects:
            o=dataset.get_object(cid)
            if o is not None:
                data=dataset.get_seg(o.t) 
                cellCoords=np.where(data==o.id)
                self.print_mn('     ----->>>  Split object '+str(o.get_name()) + " with "+str(len(cellCoords[0]))+ " voxels ")
                xmin,xmax,ymin,ymax,zmin,zmax=get_borders(data,cellCoords)
                cellShape=[1+xmax-xmin,1+ymax-ymin,1+zmax-zmin]
                markers=np.zeros(cellShape,dtype=np.uint8) #PREPARE SEEDS FOR WATERSEED
                mask=np.zeros(cellShape,dtype=np.bool)
                mask[cellCoords[0]-xmin,cellCoords[1]-ymin,cellCoords[2]-zmin]=True
                rawdata=dataset.get_raw(o.t)
                rawdata=rawdata[xmin:xmax+1,ymin:ymax+1,zmin:zmax+1]
                rawdata[np.where(mask==False)]=rawdata.max() #Remove where mask is not 
                coordinates = peak_local_max(rawdata.max()-rawdata, min_distance=min_distance,num_peaks=2) 
                self.print_mn("     ----->>>  Found "+str(len(coordinates))+ " peaks ")
                l=1
                for coord in coordinates:
                    markers[coord[0],coord[1],coord[2]]=l
                    l+=1
                labelw=watershed(rawdata,markers, mask=mask)
                
                data,newIds=apply_new_label(data,xmin,ymin,zmin,labelw,minVol=minVol)
                if len(newIds)>0:
                    dataset.del_link(o)
                    newIds.append(o.id)
                    dataset.set_seg(t,data,newIds)


        self.restart()


# -*- coding: latin-1 -*-
defaultPlugins=[]

from .watershedithNewSeeds import WatershedithNewSeeds
defaultPlugins.append(WatershedithNewSeeds())

from .ExtremeWater import ExtremeWater
defaultPlugins.append(ExtremeWater())

from .createSeeds import CreateSeeds
defaultPlugins.append(CreateSeeds())


# -*- coding: latin-1 -*-
from morphonet.plugins import MorphoPlugin


class ExtremeWater(MorphoPlugin):
    """ This plugin perform a watershed algorithm on the full image based on seeds added by a local maximum algorithm 

    Parameters
    ----------
    Gaussian_Sigma : int, default :8
        sigma parameters from the gaussian algorithm (from skimage) applied on the rawdata in otder to perform the local maximum algorithm
    MinOrMax: Dropdown
        To aplly the local minmum or local maximum algorithm (depend on the color of the bacground of the rawdata )
    Gaussian_Sigma_For Watersheed: int, default :2
        sigma parameters from the gaussian algorithm (from skimage) applied on the rawdata used in the watershed algorithm
    threshold_for_mask: int, default : 100
        a threshold parameters in which voxel will be remove on the gaussian image before performing the watershed algorithm
    RemoveBorder: Dropdown
        remove all elements which touch the border ot the segmented image 
    """

    def __init__(self): #PLUGIN DEFINITION 
        MorphoPlugin.__init__(self) 
        self.set_name("ExtremeWater")
        self.add_inputfield("gaussian_sigma",default=8)
        self.add_dropdown("MinOrMax",["min","max"])
        self.add_inputfield("gaussian_sigma_for_watershed",default=2)
        self.add_inputfield("threshold_for_mask",default=100)
       
        self.add_dropdown("RemoveBorder",["yes","no"])
        self.set_parent("Perform Segmentation")

    def process(self,t,dataset,objects): #PLUGIN EXECUTION
        if not self.start(t,dataset,objects):
            return None
        s_sigma=int(self.get_inputfield("gaussian_sigma"))
        w_sigma=int(self.get_inputfield("gaussian_sigma_for_watershed"))
        threshold=float(self.get_inputfield("threshold_for_mask"))
        MinOrMax=self.get_dropdown("MinOrMax")
        RemoveBorder=self.get_dropdown("RemoveBorder")

        from skimage.morphology import watershed,extrema
        from skimage.filters import gaussian 
        from skimage.measure import label
        import numpy as np

        membrane_image=dataset.get_raw(t)

        #Smoothing
        if s_sigma > 0.0:
            self.print_mn(" --> Perform gaussian with sigma="+str(s_sigma))
            seed_preimage=gaussian(membrane_image, sigma=s_sigma,preserve_range=True)
        else:
            seed_preimage = membrane_image
        
        #Fin Extrema
        self.print_mn(" --> Find "+MinOrMax+" extrema ")
        if MinOrMax=="min":
            local=extrema.local_minima(seed_preimage) 
        else: 
            local=extrema.local_maxima(seed_preimage) 

        #Labels Seeds
        self.print_mn(" --> Labels seeds")
        label_maxima = label(local) 

        #Perform Gaussian
        if w_sigma > 0.0:
            self.print_mn(" --> Perform gaussian with sigma="+str(w_sigma))
            seed_w_preimage=gaussian(membrane_image, sigma=w_sigma,preserve_range=True)
        else:
            seed_w_preimage = membrane_image

        #Thresholding
        mask=None
        if threshold>0.0:
            self.print_mn(" --> Perform threshold with value="+str(threshold))
            mask=seed_w_preimage>threshold


        self.print_mn(" --> Perform watershed")
        labels=watershed(seed_w_preimage, label_maxima,mask=mask)  #background ?

        #remove the element at the border
        if RemoveBorder=="yes":
            self.print_mn(" --> Remove Borders")
            borders=np.unique(np.concatenate((np.unique(labels[0,:,:]),np.unique(labels[labels.shape[0]-1,:,:]),np.unique(labels[:,0,:]),np.unique(labels[:,labels.shape[1]-1,:]),np.unique(labels[:,:,0]),np.unique(labels[:,:,labels.shape[2]-1])),axis=0))
            st=""
            for b in borders:
                st+="(labels=="+str(b)+') |'
            st=st[0:-1]
            labels[np.where(eval(st))]=0

        new_ids=np.unique(labels)
        new_ids=new_ids[new_ids!=dataset.background]
        self.print_mn(" --> Found  "+str(len(new_ids))+" labels")
       
        dataset.set_seg(t,labels,new_ids)
        self.restart()



#!/usr/bin/env python3
#
#       image.serial: read/write spatial nd images
#
#       Copyright 2006 - 2011 INRIA - CIRAD - INRA
#
#       File author(s): Jerome Chopard <jerome.chopard@sophia.inria.fr>
#                       Eric Moscardi <eric.moscardi@sophia.inria.fr>
#                       Daniel Barbeau <daniel.barbeau@sophia.inria.fr>
#
#       Distributed under the Cecill-C License.
#       See accompanying file LICENSE.txt or copy at
#           http://www.cecill.info/licences/Licence_CeCILL-C_V1-en.html
#
#       OpenAlea WebSite : http://openalea.gforge.inria.fr
################################################################################
"""
This module defines inrimage format
"""

__license__= "Cecill-C"
__revision__=" $Id$ "

import os
from os import path
import numpy as np
#from struct import calcsize,pack,unpack
import gzip
#import io
from io import BytesIO
from .spatial_image import SpatialImage

__all__ = ["read_inriheader","read_inrimage","write_inrimage"]

specific_header_keys = ("XDIM","YDIM","ZDIM",
                        "VDIM","TYPE","PIXSIZE",
                        "SCALE","CPU",
                        "VX","VY","VZ",
                        "TX","TY","TZ",
                        "#GEOMETRY")

def open_inrifile (filename) :
    """Open an inrimage file

    Manage the gz attribute
    """
    program="open_inrifile"
    if not os.path.isfile(filename) and os.path.isfile(filename+".gz"):
        filename=filename+".gz"
        #print ("%s: Warning: path to read image has been changed to %s."%(program,filename))
    if not os.path.isfile(filename) and os.path.isfile(filename+".zip"):
        filename=filename+".zip"
        #print ("%s: Warning: path to read image has been changed to %s."%(program,filename))
    if path.splitext(filename)[1] in (".gz",".zip") :
        pass
        fzip = gzip.open(filename,'rb')
        f = BytesIO(fzip.read() )
        fzip.close()
    else :
        f = open(filename,'rb')

    return f

def _read_header (f) :
    """Extract header from a stream and return it
    as a python dict
    """
    #read header string
    header = ""
    while header[-4:] != "##}\n" :
        #print("a")
        #print(f.read(256))
        header += f.read(256).decode()
        #print(header)

    #read infos in header
    prop = {}
    hstart = header.find("{\n") + 1
    hend = header.find("##}")
    infos = [gr for gr in header[hstart:hend].split("\n") \
                    if len(gr) > 0]

    #format infos
    for prop_def in infos :
        key,val = prop_def.split("=")
        prop[key] = val

    #return
    return prop

def read_inriheader (filename) :
    """Read only the header of an inrimage

    :Parameters:
     - `filename` (str) - name of the file to read
    """
    f = open_inrifile(filename)
    prop = _read_header(f)
    f.close()
    return prop

def read_inrimage (filename) :
    """Read an inrimage, either zipped or not according to extension

    :Parameters:
     - `filename` (str) - name of the file to read
    """
    f = open_inrifile(filename)

    #read header
    prop = _read_header(f)
    prop["filename"] = filename # Jonathan : 14.05.2012

    #extract usefull infos to read image
    zdim = int(prop.pop("ZDIM") )
    ydim = int(prop.pop("YDIM") )
    xdim = int(prop.pop("XDIM") )
    vdim = int(prop.pop("VDIM") )

    #if vdim != 1 :
        #   msg = "don't know how to handle vectorial pixel values"
    #   raise NotImplementedError(msg)

    #find data type
    pixsize = int(prop.pop("PIXSIZE","0").split(" ")[0])
    dtype = prop.pop("TYPE")

    if dtype == "unsigned fixed" :
        if pixsize == 0 :
            ntyp = np.dtype(np.int)
        else :
            try :
                ntyp = eval("np.dtype(np.uint%d)" % pixsize)
            except AttributeError :
                raise UserWarning("undefined pix size: %d" % pixsize)
    elif dtype == "float" :
        if pixsize == 0 :
            ntyp = np.dtype(np.float)
        else :
            try :
                ntyp = eval("np.dtype(np.float%d)" % pixsize)
            except AttributeError :
                raise UserWarning("undefined pix size: %d" % pixsize)
    else :
        msg = "unable to read that type of data : %s" % dtype
        raise UserWarning(msg)

    #read datas
    size = ntyp.itemsize * xdim * ydim * zdim * vdim
    mat = np.fromstring(f.read(size),ntyp)
    if vdim != 1 :
        mat = mat.reshape( (vdim,xdim,ydim,zdim), order="F" )
        mat = mat.transpose(1,2,3,0)
    else:
        mat = mat.reshape( (xdim,ydim,zdim), order="F" )
        #mat = mat.transpose(2,1,0)

    #create SpatialImage
    res = tuple(float(prop.pop(k) ) for k in ("VX","VY","VZ") )
    for k in ("TX","TY","TZ") :
        prop.pop(k,None)

    img = SpatialImage(mat,res,vdim,prop)

    #return
    f.close()
    return img


def write_inrimage_to_stream(stream, img):
    assert img.ndim in (3,4)

    #metadata
    info = dict(getattr(img,"info",{}) )

    #image dimensions
    if img.ndim < 4 :
        info["XDIM"],info["YDIM"],info["ZDIM"] = ("%d" % val for val in img.shape)
        info["VDIM"] = "1"
    else:
        info["XDIM"],info["YDIM"],info["ZDIM"],info["VDIM"] = ("%d" % val for val in img.shape)

    #image resolution
    res = getattr(img,"resolution",(1,1,1) )
    info["VX"],info["VY"],info["VZ"] = ("%f" % val for val in res)

    #data type
    if img.dtype == np.uint8 :
        info["TYPE"] = "unsigned fixed"
        info["PIXSIZE"] = "8 bits"
    elif img.dtype == np.uint16 :
        info["TYPE"] = "unsigned fixed"
        info["PIXSIZE"] = "16 bits"
    elif img.dtype == np.uint32 :
        info["TYPE"] = "unsigned fixed"
        info["PIXSIZE"] = "32 bits"
    elif img.dtype == np.uint64 :
        info["TYPE"] = "unsigned fixed"
        info["PIXSIZE"] = "64 bits"
    elif img.dtype == np.float32 :
        info["TYPE"] = "float"
        info["PIXSIZE"] = "32 bits"
    elif img.dtype == np.float64 :
        info["TYPE"] = "float"
        info["PIXSIZE"] = "64 bits"
    #elif img.dtype == np.float128 :
    #   info["TYPE"] = "float"
    #   info["PIXSIZE"] = "128 bits"
    else :
        msg = "unable to write that type of datas : %s" % str(img.dtype)
        raise UserWarning(msg)

    #mandatory else an error occurs when reading image
    info['#GEOMETRY']='CARTESIAN'
    info['CPU']='decm'

    #write header
    header = "#INRIMAGE-4#{\n"
    for k in specific_header_keys :#HACK pas bo to ensure order of specific headers
        try:
            header += "%s=%s\n" % (k,info[k])
        except KeyError:
            pass

    for k in set(info) - set(specific_header_keys) :
        header += "%s=%s\n" % (k,info[k])

    #fill header to be a multiple of 256
    header_size = len(header) + 4
    if (header_size % 256) > 0 :
        header += "\n" * ( 256 - header_size % 256 )

    header += "##}\n"
    stream.write(header.encode())
    if img.ndim == 3:
        stream.write(img.tostring("F"))
    elif img.ndim == 4 :
        mat = img.transpose(3,0,1,2)
        stream.write(mat.tostring("F") )
    else:
        raise Exception("Unhandled image dimension %d."%img.ndim)


def write_inrimage (filename, img) :
    """Write an inrimage zipped or not according to the extension

    .. warning:: if img is not a |SpatialImage|, default values will be used
                 for the resolution of the image

    :Parameters:
     - `img` (|SpatialImage|) - image to write
     - `filename` (str) - name of the file to read
    """
    #open stream
    zipped = ( path.splitext(filename)[1] in (".gz",".zip") )

    if zipped :
        f = gzip.GzipFile(filename, "wb")
        #f = StringIO()
    else :
        f = open(filename,'wb')

    try:
        write_inrimage_to_stream(f, img)
    except:
        # -- remove probably corrupt file--
        f.close()
        if path.exists(filename) and path.isfile(filename):
            os.remove(filename)
        raise
    else:
        f.close()



# -*- coding: latin-1 -*-
from morphonet.plugins import MorphoPlugin

class DelTemporalLink(MorphoPlugin):
    """This plugin delete any temporal links between objects

    Parameters
    ----------
    Objects: 
        It can be apply either on selected or colored objects
    """
    def __init__(self): #PLUGIN DEFINITION 
        MorphoPlugin.__init__(self) 
        self.set_name("Delete Links")
        self.set_parent("Temporal Relation")

    def process(self,t,dataset,objects): #PLUGIN EXECUTION
        if not self.start(t,dataset,objects):
            return None
        for cid in objects:
             o=dataset.get_object(cid)
             if o is not None:
                dataset.del_link(o)
                self.print_mn(" ----> remove link for object "+str(o.id)+" at "+str(o.t))
        self.restart()
        



# -*- coding: latin-1 -*-
defaultPlugins=[]

from .addTemporalLink import AddTemporalLink
defaultPlugins.append(AddTemporalLink())

from .delTemporalLink import DelTemporalLink
defaultPlugins.append(DelTemporalLink())


from .propagate import Propagate
defaultPlugins.append(Propagate())


#!/usr/bin/env python3

#
#       spatial_image: spatial nd images
#
#       Copyright 2006 INRIA - CIRAD - INRA
#
#       File author(s): Jerome Chopard <jerome.chopard@sophia.inria.fr>
#
#       Distributed under the Cecill-C License.
#       See accompanying file LICENSE.txt or copy at
#           http://www.cecill.info/licences/Licence_CeCILL-C_V1-en.html
#
#       OpenAlea WebSite : http://openalea.gforge.inria.fr
#

"""
This module create the main |SpatialImage| object
"""

__license__= "Cecill-C"
__revision__=" $Id: $ "

import numpy as np
from scipy import ndimage
import copy as cp

# -- deprecation messages --
import warnings
msg = "SpatialImage.resolution is deprecated, use SpatialImage.voxelsize"
rezexc = PendingDeprecationWarning(msg)

class SpatialImage(np.ndarray) :
    """
    Associate meta data to np.ndarray
    """
    def __new__ (cls, input_array, voxelsize = None,
             vdim = None, info = None, dtype = None, **kwargs) :
        """Instantiate a new |SpatialImage|
        
        if voxelsize is None, vdim will be used to infer space size and affect
        a voxelsize of 1 in each direction of space
        
        .. warning :: `resolution` keyword is deprecated. Use `voxelsize` instead.
        
        :Parameters:
         - `cls` - internal python
         - `input_array` (array) - data to put in the image
         - `voxelsize` (tuple of float) - spatial extension in each direction
                                           of space
         - `vdim` (int) - size of data if vector data are used
         - `info` (dict of str|any) - metainfo
        """
        #if the input_array is 2D we can reshape it to 3D.
        #~ if input_array.ndim == 2: # Jonathan
            #~ input_array = input_array.reshape( input_array.shape+(1,) ) # Jonathan
        
        #initialize datas. For some obscure reason, we want the data
        #to be F-Contiguous in the NUMPY sense. I mean, if this is not
        #respected, we will have problems when communicating with
        #C-Code... yeah, that makes so much sense (fortran-contiguous
        #to be c-readable...).
        dtype = dtype if dtype is not None else input_array.dtype
        if input_array.flags.f_contiguous :
            obj = np.asarray(input_array, dtype=dtype).view(cls)
        else :
            obj = np.asarray(input_array, dtype=dtype, order='F').view(cls)
        
        voxelsize = kwargs.get("resolution", voxelsize) #to manage transition

        if voxelsize is None :
            #~ voxelsize = (1.,) * 3
            #~ voxelsize = (1.,) * input_array.ndim # Jonathan
            try:
                voxelsize = input_array.voxelsize # Gael...
            except:
                voxelsize = (1.,) * input_array.ndim
        else :
            #~ if len(voxelsize) != 3 :
            if (input_array.ndim != 4) and (len(voxelsize) != input_array.ndim) : # Jonathan _ Compatibility with "champs_*.inr.gz" generated by Baloo & SuperBaloo
                raise ValueError("data dimension and voxelsize mismatch")
        
        obj.voxelsize = tuple(voxelsize)
        obj.vdim = vdim if vdim else 1 

        #set metadata
        if info is None :
            obj.info = {}
        else :
            obj.info = dict(info)
        
        #return
        return obj
    
    def _get_resolution(self):
        warnings.warn(rezexc)
        return self.voxelsize
    
    def _set_resolution(self, val):
        warnings.warn(rezexc)
        self.voxelsize = val
    
    resolution = property(_get_resolution, _set_resolution)
    
    @property
    def real_shape(self):
        #~ return np.multiply(self.shape[:3], self.voxelsize)
        return np.multiply(self.shape, self.voxelsize)  # Jonathan

    def invert_z_axis( self ):
        """
        invert allong 'Z' axis
        """
        self = self[:,:,::-1]

    def __array_finalize__ (self, obj) :
        if obj is None :
            return

        #assert resolution
        res = getattr(obj, 'voxelsize', None)
        if res is None :#assert vdim == 1
            res = (1.,) * len(obj.shape)

        self.voxelsize = tuple(res)

        #metadata
        self.info = dict(getattr(obj, 'info', {}) )

    def clone (self, data) :
        """Clone the current image metadata
        on the given data.

        .. warning:: vdim is defined according to self.voxelsize and data.shape

        :Parameters:
         - `data` - (array)
	
        :Returns Type: |SpatialImage|
        """
        if len(data.shape) == len(self.voxelsize) :
            vdim = 1
        elif len(data.shape) - len(self.voxelsize) == 1 :
            vdim =data.shape[-1]
        else :
            raise UserWarning("unable to handle such data dimension")

        return SpatialImage(data,self.voxelsize,vdim,self.info)

    @classmethod
    def valid_array(cls, array_like):
        return isinstance(array_like, (np.ndarray, cls)) and \
            array_like.flags.f_contiguous


def empty_image_like(spatial_image):
    array = np.zeros( spatial_image.shape, dtype=spatial_image.dtype )
    return SpatialImage(array, spatial_image.voxelsize, vdim=1)


def null_vector_field_like(spatial_image):
    array = np.zeros( list(spatial_image.shape)+[3], dtype=np.float32 )
    return SpatialImage(array, spatial_image.voxelsize, vdim=3)

def random_vector_field_like(spatial_image, smooth=0, max_=1):
    #~ if spatial_image.vdim == 1:
        #~ shape = spatial_image.shape+(3,)
    #~ else:
        #~ shape = spatial_image.shape
    shape = spatial_image.shape # Jonathan
    array = np.random.uniform(-max_, max_, shape)
    if smooth:
        array = ndimage.gaussian_filter(array, smooth)
    return SpatialImage(array, spatial_image.voxelsize, dtype=np.float32)


def checkerboard(nx=9, ny=8, nz=5, size=10, vs=(1.,1.,1.), dtype=np.uint8):
    """Creates a 3D checkerboard image with `nx` squares in width,
    `ny` squares in height and `nz` squares in depth. The length of the edge in real units
    of each square is `size`."""

    sxv, syv, szv = np.array([size]*3) / np.array(vs)
    array = np.zeros( (sxv*nx, syv*ny, szv*nz), dtype=dtype, order="F")
    typeinfo = np.iinfo(dtype)

    # -- wooo surely not the most beautiful implementation out here --
    for k in xrange(nz):
        kval = typeinfo.max if (k%2==0) else typeinfo.min
        jval = kval
        for j in xrange(ny):
            ival = jval
            for i in xrange(nx):
                array[i*sxv:i*sxv+sxv, j*syv:j*syv+syv, k*szv:k*szv+szv] = ival
                ival = typeinfo.max if (ival==typeinfo.min) else typeinfo.min
            jval = typeinfo.max if (jval==typeinfo.min) else typeinfo.min
        kval = typeinfo.max if (kval==typeinfo.min) else typeinfo.min

    return SpatialImage(array, vs, dtype=dtype)


def is2D(image):
    """
    Test if the `image` (array) is in 2D or 3D.
    Return True if 2D, False if not.
    """
    if len(image.shape) == 2 or image.shape[2] == 1:
        return True
    else:
        return False



# -*- coding: latin-1 -*-
from morphonet.plugins import MorphoPlugin

class FuseSelectedObjects(MorphoPlugin):
    """ This plugin fuse opbjects in the segmented image
   
    Parameters
    ----------
    Objects: 
        It can be apply either on selected objects or on colored objects where fusion will done by selection id
    """

    def __init__(self): #PLUGIN DEFINITION 
        MorphoPlugin.__init__(self) 
        self.set_name("Fuse")
        self.set_parent("Remove objects")

    def process(self,t,dataset,objects): #PLUGIN EXECUTION
        if not self.start(t,dataset,objects):
            return None
        import numpy as np
        times=[]
        for cid in objects: #List all time points
            o=dataset.get_object(cid)
            if o is not None and o.t not in times:
                times.append(o.t)
        times.sort() #Order Times


        for t in times:
            tofuse={}
            for cid in objects:
                o=dataset.get_object(cid)
                if o is not None and o.t==t: 
                    if o.s not in tofuse:
                        tofuse[o.s]=[]
                    tofuse[o.s].append(o.id)
            for s in tofuse:
                if len(tofuse[s])>1 : #More than one object to fuse..
                    data=dataset.get_seg(t)
                    minFuse=np.array(tofuse[s]).min()
                    cells_updated=[minFuse]
                    self.print_mn(" --> fuse objects "+str(tofuse[s])+" at "+str(t) + " in "+str(minFuse))
                    for tof in tofuse[s]:
                        if tof!=minFuse:
                            cells_updated.append(tof)
                            data[np.where(data==tof)]=minFuse
                            dataset.del_link(dataset.get_object(t,tof)) 
                    dataset.set_seg(t,data,cells_updated)
        self.restart()






# -*- coding: latin-1 -*-
import os,sys,errno
import subprocess

import numpy as np
from datetime import datetime
from morphonet.tools import _save_seg_thread,imread,imsave,isfile,copy,get_id_t,get_name,_get_param,_add_line_in_file,_read_last_line_in_file,get_info_type
from os.path import isdir,join,dirname,basename
from morphonet.tools import convert_to_OBJ,write_XML_properties,read_XML_properties,get_longid,_check_version
from threading import Thread
import threading
from http.server import HTTPServer, BaseHTTPRequestHandler
from urllib.parse import unquote
from io import BytesIO
import signal

class MorphoCuration():
    """ Curation
    """

    def __init__(self,value,date=None,active=True):
        self.active = active
        self.value=value
        self.date=date
        if self.date is None:
            self.date=datetime.now().strftime('%Y-%m-%d %H:%M:%S')


class MorphoObject():
    """ Object
    """
    def __init__(self,t,id):
        self.t =t
        self.id= id
        self.mother=None
        self.daughters=[]

    def get_name(self):
        return get_name(self.t,self.id)

    def set_mother(self,m):
        self.mother=m
        m.add_daughter(self)

    def add_daughter(self,d):
        if d not in self.daughters:
            self.daughters.append(d)

    def del_mother(self):
        if self.mother is not None:
            if self in self.mother.daughters:
                self.mother.daughters.remove(self)
        self.mother=None

    def del_daughter(self,d):
        if d in self.daughters:
            d.mother=None
            self.daughters.remove(d)

    def del_daughters(self):
        for d in self.daughters:
            d.mother=None
        self.daughters.clear()

    def del_links(self):
        self.del_mother()
        self.del_daughters()

    def nb_daughters(self):
        return len(self.daughters)

class MorphoInfo():
    """ Information that can be added in the Infos menu on the MorphoNet windows

    Parameters
    ----------
    name : string
        the name of the info
    info_type : string
        the type of the info as definied in the MorphoNet format  https://morphonet.org/help_format

    """
    def __init__(self,dataset,name,info_type):
        self.name=name
        self.dataset=dataset
        self.info_type=info_type
        self.data={}
        self.updated=True
       
    def clear(self):
        self.data.clear()

    def set(self,mo,value):
        """
        Add the new value and set all the others inactive
        """
        self.inactivate_curations(mo)
        self.add_curation(mo,value)

    def inactivate_all_curations(self):
        for mo in self.data:
            self.inactivate_curations(mo)

    def inactivate_using_daughter(self,a_daughter):
        if a_daughter.mother in self.data:
            for val in self.data[a_daughter.mother]:
                if type(val.value) == list or type(val.value) == dict:
                    for item in val.value:
                        if item == a_daughter:
                            val.value.remove(item)
                else :
                    if val.value == a_daughter:
                        self.data[a_daughter.mother].remove(val)


    def inactivate_curations(self,mo):
        if mo in self.data:
            for mc in self.data[mo]:
                mc.active=False

    def add_data(self,data):
        self.inactivate_all_curations()
        if data is None:
            return False
        if type(data) == str: #Parse Text Info as in MorphoNet
            for d in data.split('\n'):
                if not d.startswith("#") and not d.startswith(("type")):
                    dos=d.split(":")
                    if len(dos)==2:
                        mo=self.dataset.get_object(dos[0].strip())
                        self.add_curation(mo, dos[1].strip())
        else:
            self.data=data

    def add_curation(self,mo,value,date=None,active=True):
        '''
        Add a value to the info with the currente date and time
        Parameters
        ----------
        mo : MorphoObject : the cell object
        value : string  : the value
        Examples
        --------
        >>> info.add(mo,"a7.8")
        '''

        if mo not in self.data:
            self.data[mo]=[]

        if self.info_type == "time":
            if type(value) == str:
                value = self.dataset.get_object(value)

        if self.info_type == "selection":
            if type(value) == str:
                value=int(value)

        if self.info_type == "float":
            if type(value) == str:
                value=float(value)

        mc=MorphoCuration(value,date=date,active=active)
        self.data[mo].append(mc)
        self.updated=True

    def del_curation(self,mo,mc):
        if mo not in self.data:
            return False

        if mc not in self.data[mo]:
            return False

        self.data[mo].remove(mc)
        return True

    def get(self,mo):
        if mo is None:
            return None
        if mo not in self.data:
            return None
        list_info=[]
        for mc in self.data[mo]:
            if mc.active:
                list_info.append(mc.value)
        if len(list_info) == 0:
            return None
        if len(list_info)==1:
            return list_info[0]
        return list_info

    def get_curations(self,mo):
        if mo is None:
            return []
        if mo not in self.data:
            return []
        return self.data[mo]

    def get_curation(self,mo,value,date=None):
        if mo is None:
            return None
        if mo not in self.data:
            return None
        for mc in self.data[mo]:
            if value==mc.value:
                if date is None:
                    return mc
                elif mc.date==date:
                    return mc
        return None

    def get_txt(self,time_begin=-1,time_end=-1,active=True):
        Text = "#MorphoPlot" + '\n'
        Text +="#"
        if not active:
            Text += 'Curation for '
        Text +=  self.name + '\n'
        Text += "type:" + self.info_type + "\n"
        for o in self.data:
            if (time_begin == -1 or (time_begin >= 0 and o.t >= time_begin)) and (
                    time_end == -1 or (time_end >= time_begin and o.t <= time_end)):
                for mc in self.get_curations(o):
                    if mc.active == active:
                        if self.info_type == "time":
                            if type(mc.value) == dict or type(mc.value) == list:
                                for ds in mc.value:
                                    Text += o.get_name() + ':' + ds.get_name()
                                    if not active:
                                        Text+= "#"+str(mc.date)
                                    Text += '\n'
                            else :
                                Text += o.get_name() + ':' + mc.value.get_name()
                                if not active:
                                    Text += "#"+str(mc.date)
                                Text += '\n'
                        else:
                            Text += o.get_name() + ':' + str(mc.value)
                            if not active:
                                Text += str(mc.date)
                            Text += '\n'

        return Text

    def is_curated(self):
        for o in self.data:
            if len(self.get_curations(o))>1:
                return True
        return False

    def get_dict(self):
        prop={}
        for o in self.data:
            cv = o.t * 10 ** 4 + o.id
            for mc in self.get_curations(o):
                if mc.active == True:
                    if self.info_type == "time":
                        if type(mc.value) == dict or type(mc.value) == list:
                            for m in mc.value:
                                mother = m.t * 10 ** 4 + m.id
                                if m.t < o.t:
                                    if mother not in prop:
                                        prop[mother] = []
                                    prop[mother].append(cv)
                                else:  # Inverse
                                    if cv not in prop:
                                        prop[cv] = []
                                    prop[cv].append(mother)
                        else :
                            mother = mc.value.t * 10 ** 4 + mc.value.id
                            if mc.value.t < o.t:
                                if mother not in prop:
                                    prop[mother] = []
                                prop[mother].append(cv)
                            else:  # Inverse
                                if cv not in prop:
                                    prop[cv] = []
                                prop[cv].append(mother)
                    else:
                        if cv in prop:
                            if type(prop[cv])==list:
                                prop[cv].append(mc.value)
                            else:
                                prop[cv]=[prop[cv],mc.value]
                        else:
                            prop[cv]=mc.value
        return prop

    def write_curation(self,txt_filename):
        print(" --> save "+txt_filename)
        f=open(txt_filename,'w')
        f.write(self.get_txt(active=False))
        f.close()

    def read_curation(self,txt_filename):
        if os.path.isfile(txt_filename):
            f=open(txt_filename,'r')
            for line in f:
                if line.find("#")!=0 and line.find("type")==-1: 
                    p=line.find(":")
                    d=line.find("#")
                    o=self.dataset.get_object(line[:p])
                    value=line[p+1:d]
                    self.add_curation(o, value, date=line[d + 1:].strip(),active=False)
            f.close()

class Dataset():
    """Dataset class automatically created when you specify your dataset path in the seDataset function from Plot()

    Parameters
    ----------
    begin : int
        minimal time point
    end : int 
        maximal time point
    raw : string
        path to raw data file where time digits are in standard format (ex: (:03d) for 3 digits  )(accept .gz)
    segment : string
        path to segmented data file  where time digits are in standard format  (accept .gz)
    log : bool
        keep the log
    background : int
        the pixel value of the background inside the segmented image 
    xml_file : string
        path to the xml propertie files (.xml)
    memory : int
        number of time step keep in memory durig curation (if you have memeory issue, decrease this number)
    """

    def __init__(self,parent,begin=0,end=0,raw=None,segment=None,log=True,background=0,xml_file=None,memory=20):
        self.parent=parent
        self.begin=begin
        self.end=end
        self.log=log
        #List of Cells
        self.cells = {}
    
        #raw data
        self.raw=False
        self.show_raw=None
        self.raw_datas={}  #list of each rawdata time point  
        if raw is not None:
            self.raw=True
            self.raw_path=dirname(raw)+"/"
            if dirname(raw)=="":
                self.raw_path=""
            self.raw_files=basename(raw)

        #Segmentation
        self.seg_datas={}  #list of each segmented time point 
        self.segment_path=""
        self.segment_files="curated_t{:03d}.inr.gz"
        if segment is not None:
            self.segment_path=dirname(segment)+"/"
            if dirname(segment)=="":
                self.segment_path=""
            self.segment_files=basename(segment)
            if self.segment_path!="" and not isdir(self.segment_path):
                os.mkdir(self.segment_path)

        #LOG
        self.log_file = "morpho_log.txt"

        self.background=background #Background Color
            
        #DATA Management
        self.memory=memory #Memory to store dataset in Gibabytes
        self.lasT=[] #List of last time step
        self.times=[] #List of modified time point 
        
        #INFOS
        self.infos={} #For all infos Infos
        self.xml_file = xml_file #Xml Properties
        self.read_properties(self.xml_file)  #Lineage Initialisation

        self.seeds=None #To Send Centers to Unity

        #Cell to update
        self.cells_updated={}

    def print_mn(self,msg):
        """Print a string on the 3D viewer

        Parameters
        ----------
        msg : string
            your message to print 
       
        """
        self.parent.print_mn(msg)

    def save_log(self,command,exec_time):
        """Save the specitic command in the log file 

        Parameters
        ----------
        command : string
            Executed Command
        exec_time : float
            time of execution 

        Examples
        --------
        >>> dataset.save_log("fuse",date)
        """

        if self.log :
            _add_line_in_file(self.log_file,str(command)+str(exec_time.strftime("%Y-%m-%d-%H-%M-%S"))+"\n")

    def restart(self,plug):  #Apply and Restart a Curation 
        """Restart the curation mode after execution of a specific plugin

        Parameters
        ----------
        plug : MorphoPlug
            the plugin just executed

        Examples
        --------
        >>> dataset.restart(fuse)
        """

        if plug is not None:
            print(" --> Done " +str(plug.name))
            for t in self.times:
                self._save_seg(t,plug.exec_time)
        
        self.parent.restart(self.times)


    def backup(self,command,t,objects,multiple_times = False,multiple_times_list = None):
        '''
        Backup XML files and Image before performing an action
        '''
        if not isdir(".backup_morphonet"):
            os.mkdir(".backup_morphonet")
        backup_folder=join(".backup_morphonet",str(datetime.timestamp(datetime.now())))
        os.mkdir(backup_folder)

        #SAVE ACTION
        _add_line_in_file(join(backup_folder,"action.txt"), command)

        # BACKUP XML
        if self.xml_file is not None:
            os.system('cp '+self.xml_file+" "+backup_folder+" &")

        #BACKUP IMAGE
        times = []
        if multiple_times:
            times = multiple_times_list
        else :
            for cid in objects:  # List all time points
                o = self.get_object(cid)
                if o is not None and o.t not in times:
                    times.append(o.t)
        times.sort()  # Order Times
        for t in times:
            self.print_mn(" > Backup for time points : " + str(t))
            filename = join(self.segment_path, self.segment_files.format(t))
            command = 'cp '+filename+" "+backup_folder+" &"
            proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            while proc.poll() is None:
                commandResult = proc.stdout.readline()  # give output from your execution/your own message
            return_status = proc.wait()  # catch return code
            if "No such file or directory" in str(commandResult):
                #try compressed version
                command = 'cp ' + filename + ".gz " + backup_folder + " &"
                proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
                while proc.poll() is None:
                    commandResult = proc.stdout.readline()  # give output from your execution/your own message
                return_status = proc.wait()  # catch return code

    def cancel(self):
        '''
        Cancel the last action (by put the backup back)
        '''
        #Look for the last backup
        backups=[]
        for back in os.listdir(".backup_morphonet"):
            if isdir(join(".backup_morphonet",back)):
                backups.append(back)
        backups.sort()  # Order Times
        if len(backups)==0:
            print(" --> no backup found")
            return False

        last_backup=backups[len(backups)-1]
        backup_folder=join(".backup_morphonet",last_backup)

        #READ COMMAND
        selection = ""
        if isfile(join(backup_folder,"action.txt")):
            action=_read_last_line_in_file(join(backup_folder,"action.txt"))
            print(">> Cancel " + action)
            self.print_mn("Cancel "+action)
            self.save_log("CANCEL "+action,datetime.now())

            #Retrieve the list of cells
            for a in action.split(";"):
                if a.strip().startswith("ID:"):
                    objts=a[a.find('[')+1:a.find(']')].split("',")
                    for o in a[a.find('[')+1:a.find(']')].split("',"):
                        selection+=o.replace("'","")+";"

        #RESTORE XML
        if isfile(join(backup_folder,os.path.basename(self.xml_file))):
            self.read_properties(join(backup_folder,os.path.basename(self.xml_file)),reload=True)

        #RESTORE Images
        times=[]
        for t in range(self.begin,self.end+1):
            filename = join(backup_folder, self.segment_files.format(t))
            if isfile(filename):
                times.append(t)
                os.system('cp -f ' + filename + " " + self.segment_path)
                self.cells_updated[t]=None #Recompute all cells
                data=self.get_seg(t,reload=True)
                self._set_volume(data, t)

        #REMOVE BACKUP
        os.system('rm -rf '+backup_folder)

        self.parent.restart(self.times,selection=selection)

    #OBJECT ACCESS

    def get_object(self,*args):
        """Get an MorphoObject from a list of arguments (times, id, ... )

        Parameters
        ----------
        *args : list of arugemnts 
            the arguments which define the object, with at least 1 argument (object id with time =0 )

        Return 
        ----------
        MorphoObject class 

        Examples
        --------
        >>> dataset.get_object(1,2)
        """
        t=0
        id=None
        s=None #Selection
        tab=args
        if len(args) == 1:
            tab=args[0].split(",")

        if len(tab) == 1:
            id = int(tab[0])
        elif len(tab) >= 2:
            t = int(tab[0])
            id = int(tab[1])
        if len(tab) >= 3:
            s=int(tab[2])


        if id is None:
            print(" Wrong parsing  " + str(args[0]))
            return None

        if t not in self.cells:
            self.cells[t] = {}
        if id not in self.cells[t]: #CREATION
            self.cells[t][id] = MorphoObject(t,id)

        if s is not None:
            self.cells[t][id].s=s

        return self.cells[t][id]

    ##### DATA ACCESS 

    def _set_last(self,t):
        if t in self.lasT:
            self.lasT.remove(t)
        self.lasT.append(t)
        if t not in self.seg_datas:
            if self._get_data_size()>self.memory*10**9:
                remove_t=self.lasT.pop(0)
                if remove_t in self.seg_datas:
                    del self.seg_datas[remove_t]
                if remove_t in self.raw_datas:
                    del self.raw_datas[remove_t] 

    def _get_data_size(self):
        sif=0
        for t in self.seg_datas:
            if self.seg_datas[t] is not None:
                sif+=self.seg_datas[t].nbytes
        return sif

    def _set_volume(self,data,t):
        #Compute new Volumes
        inf=self.get_info('cell_volume',info_type="float")
        factor=4 #Computational Factor to reduce time computation
        dataResize=data[::factor,::factor,::factor]
        cells=np.unique(dataResize)
        cells=cells[cells!=self.background]
        for c in cells:
            newV=len(np.where(dataResize==c)[0])*(factor*factor*factor)
            o=self.get_object(t,c)
            if inf.get_curation(o,newV) is None:
                inf.inactivate_curations(c)
                inf.add_curation(o,newV)
        del dataResize

    def set_seg(self,t,data,cells_updated=None):
        """Define the segmented data at a specitic time point

        Parameters
        ----------
        t : int
            the time point 
        data : numpy matrix
            the segmented image
        cells_updated (optional): list
            list of cell just udpated by the plugin (in order to compute faster)

        Examples
        --------
        >>> dataset.set_seg(1,data)
        """

        self.seg_datas[t]=data
        if t not in self.times:
            self.times.append(t)

        self.update_cells(t,cells_updated)

    def update_cells(self,t,cells_updated):
        """Updated list of modified cells given a specific time step

        Parameters
        ----------
        t : int
            the time point 
        cells_updated : list or MorphoObject
            list of cell just udpated by the plugin (in order to compute faster)

        Examples
        --------
        >>> dataset.update_cells(1,[o,d])
        """

        if t not in self.cells_updated:
            self.cells_updated[t]=[]

        if cells_updated is not None:
            for c in cells_updated:
                if c not in self.cells_updated[t]:
                    self.cells_updated[t].append(c)

    def _save_seg(self,t,exec_time,data=None):
        if data is None:
            data=self.seg_datas[t]
        else:
            self.seg_datas[t]=data
        self._set_volume(data,t)

        sst=_save_seg_thread(self.segment_path,self.segment_files,data,t,exec_time)
        sst.start()

    def get_raw(self,t):
        """Get the rawdata data at a specitic time point

        Parameters
        ----------
        t : int
            the time point 
        Return
        ----------
        numpy matrix
            the raw data

        Examples
        --------
        >>> dataset.get_raw(1)
        """
        if self.raw_path is None:
            print(" Miss raw path")
            return None
        filename=join(self.raw_path,self.raw_files.format(t))
        if not os.path.isfile(filename):
            print(" Miss raw file "+filename)
            return None
        if t not in self.raw_datas:
            self.raw_datas[t]=imread(join(self.raw_path,self.raw_files.format(t)))
        self._set_last(t)  # Define the time step as used
        return self.raw_datas[t]

    def get_seg(self,t,reload=False):
        """Get the segmented data at a specitic time point

        Parameters
        ----------
        t : int
            the time point 

        Return
        ----------
        numpy matrix
            the segmented image

        Examples
        --------
        >>> dataset.get_seg(1)
        """
        self._set_last(t) #Define the time step as used
        if t not in self.seg_datas or reload:
            self.seg_datas[t]=None
            if isfile(join(self.segment_path,self.segment_files.format(t))):
                self.seg_datas[t]=imread(join(self.segment_path,self.segment_files.format(t)))
        return self.seg_datas[t]

    def get_all_seg(self,reload=False):
        """Get the segmented data for all time points

        Return
        ----------
        tab of numpy matrix
            the segmented images list

        Examples
        --------
        >>> dataset.get_seg(1)
        """
        for t in self.times:
            if t not in self.seg_datas or reload:
                self.seg_datas[t] = None
                if isfile(join(self.segment_path, self.segment_files.format(t))):
                    self.seg_datas[t] = imread(join(self.segment_path, self.segment_files.format(t)))
        return self.seg_datas

    def get_center(self,data): #Calculate the center of a dataset
        """Get the barycnetr of an matrix passed in argument 

        Parameters
        ----------
        data : numpy matrix
            the 3D image (could be segmented or rawdata) 

        Return
        ----------
        list of coordinates 
            the barycenter of the image 

        Examples
        --------
        >>> center=dataset.get_center(seg)
        """

        return [np.round(data.shape[0]/2),np.round(data.shape[1]/2),np.round(data.shape[2]/2)]

    def add_seed(self,seed):
        """Add a seed in the seed list

        Parameters
        ----------
        seed : numpy array
            the coordinate of a seed 


        Examples
        --------
        >>> dataset.add_seed(np.int32([23,34,45]))
        """

        if self.seeds is None:
            self.seeds=[]
        self.seeds.append(seed)

    def get_seeds(self):
        """Return the list of seeds as string

        Examples
        --------
        >>> seeds=mc.get_seeds()
        """

        if self.seeds is None or len(self.seeds)==0:
            return None
        strseed=""
        for s in self.seeds:
            strseed+=str(s[0])+","+str(s[1])+","+str(s[2])+";"
        self.seeds=None #Reinitializeation
        return strseed[0:-1]

    ##### LINEAGE FUNCTIONS

    def read_properties(self,filename,reload=False):
        if filename is not None :
            prop_path=os.path.dirname(filename)
            properties=read_XML_properties(filename)
            if properties is not None:
                for info_name in properties:
                    if info_name!="all_cells":
                        print( " ----> found "+info_name)
                        prop=properties[info_name]
                        if prop is not None:
                            info_type=get_info_type(info_name)
                            if info_name.find("selection_")==0:
                                info_name=info_name.replace("selection_","")
                            if info_type is None:
                                info_type="string"
                            if type(prop) == list:
                                info_type = "selection"
                            inf=self.get_info(info_name,info_type=info_type,reload=reload)
                            if type(prop)==list: #List of Cells
                                for idl in prop:
                                    t, c = get_id_t(idl)
                                    mo = self.get_object(get_name(t, c))
                                    inf.add_curation(mo,1)
                            else:#DICT
                                for idl in prop:
                                    t,c=get_id_t(idl)
                                    mo = self.get_object(get_name(t,c))
                                    if info_type=='time':
                                        daughters=[]
                                        for daughter in prop[idl]:
                                            td,d=get_id_t(daughter)
                                            do = self.get_object(get_name(td,d))
                                            do.set_mother(mo)
                                            daughters.append(do)
                                        inf.add_curation(mo,daughters)
                                    else:
                                        if type(prop[idl])==list:
                                            for elt in prop[idl]:
                                                inf.add_curation(mo , elt)
                                        else:
                                            inf.add_curation(mo , prop[idl] )


                            inf.read_curation(join(prop_path,info_name+".log")) #READ CURATION FILE

    def save(self):
        '''
        Save the properties
        '''
        self.write_properties(self.xml_file)

    def write_properties(self,filename):
        if filename is not None:
            properties={}
            for info_name in self.infos:
                inf=self.infos[info_name]
                info_name_w=info_name
                if inf.info_type=="selection" and info_name.find("selection_")==-1:
                    info_name_w="selection_"+info_name
                properties[info_name_w]=inf.get_dict()
            properties["all_cells"]=[] #Add ALL CELLS
            inf=self.get_info("cell_volume")
            for o in inf.data:
                properties["all_cells"].append(get_longid(o.t,o.id))
            write_XML_properties(properties,filename)

    def get_info(self,info_name,info_type=None,reload=False,create=True):
        '''
        Return the info for the dataset
        '''
        if info_type is None:
            info_type=get_info_type(info_name)
            if info_type is None:
                info_type="string"

        if reload and info_name in self.infos: #Clear Info
            self.infos[info_name].clear()

        if info_name not in self.infos and create:  #Create a new one
            self.infos[info_name]=MorphoInfo(self,info_name,info_type)

        if info_name not in self.infos:
            return None
        return self.infos[info_name]

    ################## TEMPORAL FUNCTIONS 

    def _get_at(self,objects,t):
        cells=[]
        for cid in objects:
            o=self.get_object(cid)
            if o is not None and o.t==t:
                    cells.append(o)
        return cells
    
    def add_link(self,da,mo):
        """Create a temporal link in the lineage

        Parameters
        ----------
        da : MorphoObject
            the daughter cell 
        mo : MorphoObject
            the mother cell 


        Examples
        --------
        >>> mc.add_link(da,mo)
        """
        da.set_mother(mo)
        inf=self.get_info("cell_lineage",info_type="time")
        inf.add_curation(da,mo)

    def del_link(self,o): #We remove all links correspond to a cells
        """Remove all temporal relations for a sepcific in the lineage

        Parameters
        ----------
        o : MorphoObject
            the cell 

        Examples
        --------
        >>> mc.del_link(o)
        """

        print(" > Delete links for objects : "+o.get_name())
        inf = self.get_info("cell_lineage", info_type="time")
        #f = open("prev_info.txt","w")
        #f.write(inf.get_txt())
        #f.close()
        #remove currations
        for mc in inf.get_curations(o.mother):
            if mc.value==o:
                print("deleted link : "+str(o.get_name())+" -> "+o.mother.get_name())
                mc.active=False

        for d in o.daughters:
            for mc in inf.get_curations(d):
                if mc.value == o:

                    mc.active = False

        inf.inactivate_using_daughter(o)
        #remove hard links (originals)
        o.del_links()
        #
        inf.inactivate_curations(o)

        #inf is updated, so write
        #line_infos = inf.get_dict()
        #f2 = open("new_info.txt","w")
        #f2.write(inf.get_txt())
        #f2.close()
        #f3 = open("new_info_dict.txt","w")
        #for line in line_infos:
        #    f3.write(str(line)+"\n")
        #f3.close()
        inf.updated = True
        #self.save()


#****************************************************************** MORPHONET SERVER


class _MorphoServer(Thread):
    def __init__(self,ploti,todo,host="",port=9875):
        Thread.__init__(self) 
        self.ploti=ploti
        self.todo=todo
        self.host=host
        self.port=port
        self.server_address = (self.host, self.port)
        self.available = threading.Event() #For Post Waiting function
        self.lock = threading.Event()
        self.lock.set()

    def run(self): #START FUNCTION
        #print("Run server Localhost on the port ", self.port)
        if self.todo=="send":
            handler = _MorphoSendHandler(self.ploti,self)
        else: #recieve
            handler = _MorphoRecieveHandler(self.ploti,self)

        self.httpd = HTTPServer(self.server_address, handler)
        self.httpd.serve_forever()

    def reset(self):
        self.obj = None  
        self.cmd=None  
        self.available = threading.Event() #Create a new watiing process for the next post request
        self.lock.set() #Free the possibility to have a new command

    def wait(self):  #Wait free request to plot (endd of others requests)
        self.lock.wait()

    def post(self,cmd,obj): #Prepare a command to post
        self.lock = threading.Event() #LOCK THE OTHER COMMAND
        self.available.set() 
        self.cmd=cmd
        self.obj=obj
        
    def stop(self):
        self.lock.set()
        self.available.set()
        self.httpd.shutdown()

class _MorphoSendHandler(BaseHTTPRequestHandler):
 
    def __init__(self, ploti,ms):
        self.ploti = ploti
        self.ms=ms
     
    def __call__(self, *args, **kwargs): #Handle a request 
        super().__init__(*args, **kwargs)

    def _set_headers(self):
        self.send_response(200)
        self.send_header("Access-Control-Allow-Origin", "*") #To accept request from morphonet
        self.end_headers()

    def do_GET(self): #NOT USED
        self._set_headers()
        self.wfile.write(b'OK')

    def do_POST(self):
        self.ms.available.wait() #Wait the commnand available
        self._set_headers()
        #print("-> SEND for "+self.ms.cmd)#+"->"+str(self.ms.obj))
        content_length = int(self.headers['Content-Length']) 
        command =self.rfile.read(content_length)
        response = BytesIO()
        #print(command)
        response.write(bytes(self.ms.cmd, 'utf-8'))
        response.write(b';') #ALWAYS ADD A SEPARATOR
        if self.ms.obj is not None:
            if  self.ms.cmd.find("RAW")==0:
                response.write(self.ms.obj)
            else :
                response.write(bytes(self.ms.obj, 'utf-8'))
        self.wfile.write(response.getvalue())
        self.ms.cmd=""
        self.ms.obj=None
        self.ms.reset() #FREE FOR OTHERS COMMAND

    def log_message(self, format, *args):
        return

class _MorphoRecieveHandler(BaseHTTPRequestHandler):
 
    def __init__(self, ploti,ms):
        self.ploti = ploti
        self.ms=ms
     
    def __call__(self, *args, **kwargs): #Handle a request 
        super().__init__(*args, **kwargs)

    def _set_headers(self):
        self.send_response(200)
        self.send_header("Access-Control-Allow-Origin", "*") #To accept request from morphonet
        self.end_headers()

    def do_GET(self): #NOT USED
        self._set_headers()
        self.wfile.write(b'OK')

    def do_POST(self):
        self._set_headers()
        response = BytesIO() #Read 
        content_length = int(self.headers['Content-Length']) # <--- Gets the size of data
        command =self.rfile.read(content_length)
        action=_get_param(command,"action")
        current_time=int(_get_param(command,"time"))
        objects=_get_param(command,"objects").split(";")
        #print("action="+action)
        if action=="showraw":
            self.ploti.plot_raws(current_time)
        elif action=="upload":
            self.ploti.upload(objects[0],2)
        elif action == "cancel":
            self.ploti.cancel()
        elif action=="reload_infos":
            self.ploti.reload_infos()
        elif action=="create_curation":
            self.ploti.curate_info(_get_param(command,"info"),_get_param(command,"objects"),_get_param(command,"value"),_get_param(command,"date"))
        elif action=="delete_curation":
            self.ploti.delete_curate_info(_get_param(command,"info"),_get_param(command,"objects"),_get_param(command,"value"),_get_param(command,"date"))
        elif action=="delete_curation_value":
            self.ploti.delete_curate_info_using_value(_get_param(command,"info"),_get_param(command,"objects"),_get_param(command,"value")) 
        elif action=="create_info_unity":
            self.ploti.create_info_from_unity(_get_param(command,"name"),_get_param(command,"datatype"),_get_param(command,"infos"))
        elif action=="delete_info_unity":
            self.ploti.delete_info_from_unity(_get_param(command,"info"))
        elif action=="delete_selection":
            self.ploti.delete_selection_from_unity(_get_param(command,"info"),_get_param(command,"selection"))
        else:
            actions=unquote(str(command.decode('utf-8'))).split("&")
            for plug in self.ploti.plugins:
                if plug._cmd()==action: #print(" Found Plugin "+plug().cmd())
                    ifo=0 
                    for tf in plug.inputfields:
                        plug._set_inputfield(tf,actions[3+ifo][actions[3+ifo].index("=")+1:])
                        ifo+=1
 
                    for dd in plug.dropdowns:
                        plug._set_dropdown(dd,actions[3+ifo][actions[3+ifo].index("=")+1:])
                        ifo+=1
                    for cd in plug.coordinates:
                        plug._set_coordinates(cd,actions[3+ifo][actions[3+ifo].index("=")+1:])
                        ifo+=1

                    plug.process(current_time,self.ploti.dataset,objects)

        response.write(bytes("DONE", 'utf-8'))
        self.wfile.write(response.getvalue())

    def log_message(self, format, *args):
        return
    
class Plot:#Main function to initalize the plot mode
    """Plot data onto the 3D viewer of the MorphoNet Window.

    Parameters (mostly for debuging )
    ----------
    log : bool
        keep the log
    start_browser : bool
        automatically start the browser when plot initliaze
    port : int
        port number to communicate with the MorphoNet Window. 
    
    Returns
    -------
    MorphoPlot
        return an object of morphonet which will allow you to send data to the MorphoNet Window.


    Examples
    --------
    >>> import morphonet
    >>> mn=morphonet.Plot()

    """

    def __init__(self,log=True,start_browser=True,port_send=9875,port_recieve=9876,clear_backup=True,clear_temp=True):
        
        _check_version()

        self.server_send=_MorphoServer(self,"send",port=port_send) #Instantiate the local MorphoNet server
        self.server_send.start() #

        self.server_recieve=_MorphoServer(self,"recieve",port=port_recieve) #Instantiate the local MorphoNet server 
        self.server_recieve.start() 
    

        if start_browser :
            self.show_browser()
        self.plugins=[]
        self.log=log        

        signal.signal(signal.SIGINT, self._receive_signal)
        self.clear_temp=clear_temp
        if self.clear_temp:
            self._clear_temp()

        if clear_backup:
            self._clear_backup()

    #########################################  SERVER COMMUNICATION

    def write_info(self,txt_filename,data):
        print(" --> save "+txt_filename)
        f=open(txt_filename,'w')
        f.write(data)
        f.close()

    def connect(self, login,passwd): #Need to be connected to be upload on MorphoNet 
        """Connect to the MorphoNet server

        In order to directly upload data to the MorphoNet server, you have to enter your MorphoNet credentials

        Parameters
        ----------
        login : string
            your login in MorphoNet
        passwd : string
            your password in MorphoNet

        Examples
        --------
        >>> import morphonet
        >>> mc=morphonet.Plot()
        >>> mc.connect("mylogin","mypassword")
        """
        import morphonet 
        self.mn=morphonet.Net(login,passwd)
     
    def print_mn(self,msg):
        """Print a string on the 3D viewer

        Parameters
        ----------
        msg : string
            your message to print 
       
        Examples
        --------
        >>> mc=print_mn("Hello")
        """
        if msg!="DONE":
            print(msg)
        self.send("MSG",msg)

    def send(self,cmd,obj=None):
        """ Send a command to the 3D viewer

        Examples
        --------
        >>> mc.send("hello")
        """
        self.server_send.wait() #Wait the commnand available
        if cmd is not None:
            cmd=cmd.replace(" ","%20")
        if obj is not None:
            if type(obj)==str:
                obj=obj.replace(" ","%20")
        self.server_send.post(cmd,obj)

    def quit(self):
        """ Stop communication between the browser 3D viewer and python

        Examples
        --------
        >>> mc.quit()
        """
        self.server_send.stop() #Shut down the server
        self.server_recieve.stop() #Shut down the server

    def upload(self,dataname,upload_factor=2):
        """Create the dataset on MorphoNet server and upload data

        Parameters
        ----------
        dataname : string
            Name of the new dataset on the server
        upload_factor : float
            the scaling attached to the dataset to match the raw data

        Examples
        --------
        >>> ...after starting MorphoPlot and curating the data
        >>> mc.upload("new dataset name",1)
        """
        print("---->>> Upload dataset "+dataname)
        self.mn.create_dataset(dataname,minTime=self.dataset.begin,maxTime=self.dataset.end)
        for t in range(self.dataset.begin,self.dataset.end+1):
            data=self.dataset.get_seg(t)
            if data is not None:
                obj=convert_to_OBJ(data,t,background=self.dataset.background,factor=upload_factor,cells_updated=None,path_write=self.temp_path)
                self.mn.upload_mesh(t,obj)
        #TODO add Infos
        print("---->>>  Uploading done")

    def show_browser(self): 
        """ Start Mozilla Firefox browser and open morphoplot page
        
        Examples
        --------
        >>> mc.show_browser()
        """
        import webbrowser
        from morphonet import url
        print(" --> open "+url)
        try:
            webbrowser.get('firefox').open_new_tab("http://"+url+'/morphoplot')
        except Exception as e:
            print("Firefox error: " % e)
            quit()

    def curate(self): #START UPLOAD AND WAIT FOR ANNOTATION
        """ Start sending data to the browser 3D viewer, then wait for annotation from the browser

        Examples
        --------
        >>> mc=morphonet.Plot(start_browser=False)
        >>> mc.set_dataset(...)
        >>> mc.curate()
        """
        self.print_mn("Wait for the MorphoNet Windows")
        self.send("START_"+str(self.dataset.begin)+"_"+str(self.dataset.end))
        self.set_default_plugins()  #Initialise Default set of plugins
        self.plot_meshes()
        self.plot_infos()
        self.plot_infos_currated()
        self._reset_infos()
        self.print_mn("DONE")

    def restart(self,times,selection=None):
        self.dataset.write_properties(self.dataset.xml_file)

        if times is not None: #PLOT MESHES
            self.plot_meshes(times)

        if self.dataset.show_raw is not None: #PLOT RAWDATAS
            self.plot_raw(self.dataset.show_raw)

        self.plot_seeds(self.dataset.get_seeds()) #PLOT SEEDS
 
        self.plot_infos() #PLOT ALL INFOS

        self.plot_selection(selection) #PLOT SELECTION

        self._reset_infos()
        self.print_mn("DONE")

    def cancel(self):
        '''
        Cancel last action -> retrieve last backup
        '''
        self.dataset.cancel()

    #########################################  DATASET

    def set_dataset(self,begin=0,end=0,raw=None,segment=None,background=0,xml_file=None,factor=4,raw_factor=4,memory=20):
        """ Define a dataset to curate
        
        Parameters
        ----------
        begin : int
            minimal time point
        end : int 
            maximal time point
        raw : string
            path to raw data file where time digits are in standard format (ex: (:03d) for 3 digits  )(accept .gz)
        segment : string
            path to segmented data file  where time digits are in standard format  (accept .gz)
        background : int
            the pixel value of the background inside the segmented image 
        xml_file : string
            path to the xml propertie files (.xml)
        factor : int
            reduction factor when meshes are calculated and send to the MorphoNet window
        raw_factor : int
            raw data reduction factor
        memory : int
            number of time step keep in memory durig curation (if you have memeory issue, decrease this number)

        Examples
        --------
        >>> ...after connection
        >>> mc.set_dataset(self,begin=0,end=10,raw=path/to/name_t(:03d).inr.gz,segment=path/to/segmenteddata_t(:03d).inr.gz,xml_file=path/to/properties_file.xml)
        """
        self.dataset=Dataset(self,begin,end,raw=raw,segment=segment,log=self.log,background=background,xml_file=xml_file,memory=memory)
        self.center=None
        self.factor=factor #Reduce factor to compute the obj
        self.raw_factor=raw_factor #Reduction factor

        #Temporary folder
        self.temp_path=".temp_morphonet_"+str(os.path.basename(segment))
        if self.temp_path!="" and not isdir(self.temp_path):
            os.mkdir(self.temp_path)

    def save(self):
        '''
        Write properties
        '''
        self.dataset.save()

    ######################################### PLUGINS

    def add_plugin(self,plug):
        """ Add a python plugin to be import in the MorphoNet Window
        
        Parameters
        ----------
        plugin : MorphoPlugin
            A plugin instance

        Examples
        --------
        >>> from plugins.MARS import MARS
        >>> mc.add_plugin(MARS())
        """
        if plug not in self.plugins:
            self.plugins.append(plug)
            self._create_plugin(plug)

    def _create_plugin(self,plug):
        """ Create the plugin in the MorphoNet Window
        
        Parameters
        ----------
        plugin : MorphoPlugin
            A plugin instance
    
        """
        print(" --> create Plugin "+plug.name)
        self.send("BTN",plug._get_btn())
        
    def set_default_plugins(self):
        """ Load the default plugins to the 3D viewer

        Examples
        --------
        >>> mc.set_default_plugins()
        """
        from morphonet.plugins import defaultPlugins
        for plug in defaultPlugins:
            self.add_plugin(plug)

    ######################################### RAWIMAGES 
   
    def plot_raws(self,t):
        """ Enable the possibility to plot raw images to the browser for a specified time point ? 
        
        Parameters
        ----------
        t : int
            time point to display raw images from

        Examples
        --------
        >>> mc.plot_raws(1)
        """
        if self.dataset.raw:
            if self.dataset.show_raw is None or self.dataset.show_raw!=t:
                self.dataset.show_raw=t
                self.restart(None)         

    def plot_raw(self,t):
        """ Compute and send raw images to the browser for a specified time point
        
        Parameters
        ----------
        t : int
            time point to display raw images from

        Examples
        --------
        >>> mc.plot_raw(1)
        """
        if self.dataset.raw:
            print(" --> Send rawdatas at "+str(t))
            rawdata=self.dataset.get_raw(t)
            if rawdata is not None:
                new_shape=np.uint16(np.floor(np.array(rawdata.shape)/self.raw_factor)*self.raw_factor) #To avoid shifting issue
                rawdata=rawdata[0:new_shape[0],0:new_shape[1],0:new_shape[2]]
                factor_data=rawdata[::self.raw_factor,::self.raw_factor,::self.raw_factor]
                bdata=np.uint8(np.float32(np.iinfo(np.uint8).max)*factor_data/factor_data.max()).tobytes(order="F")
                if self.center is None:
                    self.center=self.dataset.get_center(rawdata)
                cmd="RAW_"+str(t)+"_"+str(rawdata.shape[0])+"_"+str(rawdata.shape[1])+"_"+str(rawdata.shape[2])+"_"+str(self.raw_factor)+"_"+self._get_centerText()
                self.send(cmd,bdata)

    ######################################### ADDD CENTERS
    
    def plot_seeds(self,seeds):
        """ Plot the cell centers to the browser
        
        Parameters
        ----------
        seeds : string
            the centers of the cells

        Examples
        --------
        >>> mc.plot_seeds(seed_info)
        """
        if seeds is not None and seeds!="":
            self.send("SEEDS",seeds)

    def _get_centerText(self):
        if self.center is not None:
            return str(int(round(self.center[0])))+"_"+str(int(round(self.center[1])))+"_"+str(int(round(self.center[2])))
        return "0_0_0"

    ######################################### PRIMITIVES  
    
    def add_primitive(self,name,obj): 
        """ Add a primitive using specified content with the specified name to the browser
        
        Parameters
        ----------
        name : string
            the name of the primitive
        obj : bytes
            content of the primitive (3D data)

        Examples
        --------
        >>> #Specify a file on the hard drive by path, with rights
        >>> f = open(filepath,"r+")
        >>> #load content of file inside variable
        >>> content = f.read()    
        >>> mc.add_primitive("primitive name",content)
        >>> f.close()
        """
        self.send("PRIM_"+str(name),obj)

    ######################################### INFOS 
    
    def _reset_infos(self):
        """
            Reset the updated of all infos 
        """
        if self.dataset.infos is not None:
            for info_name in self.dataset.infos: 
                inf=self.get_info(info_name)
                inf.updated=False

    def plot_infos(self):

        """ Plot all the infos of the datasset
        """

        if self.dataset.infos is not None:
            for info_name in self.dataset.infos: 
                self.plot_info(self.get_info(info_name))

    def plot_info(self,info): #PLOT INFO (CORRESPONDENCAE)
        """ Send the specified informations with the specified name to browser
        
        Parameters
        ----------
        info : Info Class
           the info to plot   

        Examples
        --------
        >>> my_info=mc.get_info("Cell Name")
        >>> mc.plot_infos(my_info)
        """

        
        if info is None:
            return 
        if info.updated:
            print(" --> plot "+info.name)
            info_text=info.get_txt(time_begin=self.dataset.begin,time_end=self.dataset.end)
            self.send("INFO_"+info.name,info_text)
            
    def plot_infos_currated(self):

        """ Plot all the curation for all the infos of the datasset
        """

        if self.dataset.infos is not None:
            for info_name in self.dataset.infos: 
                self.plot_info_currated(self.get_info(info_name))

    def plot_info_currated(self,info): 
        """ Send the specified currattion for the informations with the specified name to browser
        
        Parameters
        ----------
        info : Info Class
           the info to plot   
        """

        
        if info is None:
            return

        if info.is_curated():
            curation_txt = info.get_txt(time_begin=self.dataset.begin, time_end=self.dataset.end,active=False)
            self.send("CUR_"+info.name,curation_txt)

    def get_infos(self):
        """ Return all the informations associated to the dataset
        """
        return self.dataset.infos

    def get_info(self,info_name):
        """ Return the info associated to the dataset
        
        Parameters
        ----------
        info_name : string
           the name of the info

        return info : Class info 
            return an object of info 

       
        Examples
        --------  
        >>> my_info=mc.get_info("Cell Name")
        >>> my_info.get_txt() #return the text file
        """
        if info_name in self.dataset.infos:
            return self.dataset.infos[info_name]
        return None

    def create_info(self,info_name,info_type,data=None):
        """ Create an info associated to the dataset
        
        Parameters
        ----------
        info_name : string
           the name of the info
        info_type
            the type of the info (float,string, etc.. ) in string
        data (optional) : List<string> or info as in MorphoNet
            information content as a list of all lines 

        Examples
        --------  
        >>> info=mc.create_info("Cell Name","string")
        >>> info.set(el,"a7.8")
        """
        inf=self.dataset.get_info(info_name,info_type=info_type,reload=False)
        if data is not None:
            inf.add_data(data)
        return inf

    def delete_info(self,info_name):
        """ delete an info associated to the dataset
        
        Parameters
        ----------
        info_name : string
           the name of the info

        Examples
        --------  
        >>> info=mc.delete_info("Cell Name")
        >>> info.set(el,"a7.8")
        """
        if info_name in self.dataset.infos:
            self.dataset.infos.remove(info_name)

    def set_info_type(self,info_name,info_type):
        """ Change or specify the type of an info associated to the dataset
            The info can be created directly in python or load in the XML file

        Parameters
        ----------
        info_name : string
          the name of the info
        info_type
           the type of the info (float,string, etc.. )  in string

        Return True if the changement is affected

        Examples
        --------
        >>> mc.set_info_type("ThisField","selection")
        """
        infor=self.get_info(info_name)
        if infor is None:
            return False
        infor.info_type=info_type
        return True

    def reload_infos(self):
        self.plot_infos()
        self.plot_infos_currated()

    def curate_info(self,info_name,k,v,d):
        """ Apply the curration value of a specific object for the info name 
        
        Parameters
        ----------
        info_name : string
           the name of the info
        k : string
            object to curate
        v : string
            value of curation
        d : string
            date of curation  
        """
        print(" curate_info " + info_name)
        info=self.get_info(info_name)
        o=self.dataset.get_object(k)
        info.add_curation(o,v,date=d)
        self.restart(None)

    def delete_curate_info(self,info_name,k,v,d):
        """ Delete the curration value of a specific object for the info name 
        
        Parameters
        ----------
        info_name : string
           the name of the info
        k : string
            object to curate
        v : string
            value of curation
        d : string
            date of curation  
        """
        info=self.get_info(info_name)
        o = self.dataset.get_object(k)
        #o=info._get_object(MorphoObject(k))
        if not info.del_curation(o,v,d):
            print(" Error during the deletion of the curation ")
        self.restart(None)

    def delete_curate_info_using_value(self,info_name,k,v):
        """ Delete the curration value of a specific object for the info name using the value
        
        Parameters
        ----------
        info_name : string
           the name of the info
        k : string
            object to curate
        v : string
            value of curation
        """
        info=self.get_info(info_name)
        o = self.dataset.get_object(k)
        #o=info._get_object(MorphoObject(k))
        if not info.del_curation_using_value(o,v):
            print(" Error during the deletion of the curation ")
        self.restart(None)

    def create_info_from_unity(self,info_name,datatype,data):
        """ Create or Update info when receiving data from unity
        
        Parameters
        ----------
        info_name : string
           the name of the info
        datatype : string
            info type
        data : string
            data to write in info file
        """
        self.create_info(info_name,datatype,data)
        self.restart(None)

    def delete_info_from_unity(self,info_name):
        """ Create or Update info when receiving data from unity
        
        Parameters
        ----------
        info_name : string
           the name of the info
        datatype : string
            info type
        data : string
            data to write in info file
        """
        print(" --> delete info "+info_name)
        info=self.get_info(info_name)
        if info is not None:
            info.clear()
            del self.dataset.infos[info_name]
            self.restart(None)

    def delete_selection_from_unity(self,info_name,selection_number):
        """ Delete info when receiving data from unity
        
        Parameters
        ----------
        info_name : string
           the name of the info
        datatype : string
            info type
        data : string
            data to write in info file
        """
        self.delete_info(info_name)
        self.restart(None)

    #########################################  SELECTION

    def plot_selection(self,selection):
        '''
        Plot selection (list of objects separated by ;)
        '''
        if selection is not None:
            print(" --> plot selection "+selection)
            self.send("SELECT",str(selection))

    #########################################  MESH
    def _get_mesh(self,t,data):
        if self.center is None:
            self.center=self.dataset.get_center(data)
        if t not in self.dataset.cells_updated:
            self.dataset.cells_updated[t]=[]
        obj=convert_to_OBJ(data,t,background=self.dataset.background,factor=self.factor,center=self.center,cells_updated=self.dataset.cells_updated[t],path_write=self.temp_path) #Create the OBJ
        self.dataset.cells_updated[t]=[]
        return obj  
    
    def plot_mesh(self,t): #UPLOAD DITECLTY THE OBJ TIME POINT IN UNITY
        """ Send the 3D files for the specified time point to browser and display the mesh 
        
        Parameters
        ----------
        t : int
            the time point to display in browser

        Examples
        --------
        >>> mc.plot_mesh(1)
        """
        obj=""
        data=self.dataset.get_seg(t)
        if data is not None:
            print(" --> Send mesh at "+str(t))
            self.dataset._set_volume(data,t) #Update Volumes
            obj=self._get_mesh(t,data)
        self.send("LOAD_"+str(t),obj)

    def plot_at(self,t,obj):#PLOT DIRECTLY THE OBJ PASS IN ARGUMENT
        """ Plot the specified 3D data to the specified time point inside the browser
        
        Parameters
        ----------
        t : int
            the time point to display in browser
        obj : bytes
            the 3d data

        Examples
        --------
        >>> #Specify a file on the hard drive by path, with rights
        >>> f = open(filepath,"r+")
        >>> #load content of file inside variable
        >>> content = f.read()    
        >>> mc.plot_at(1,content)
        >>> f.close()
        """
        self.send("LOAD_"+str(t),obj)

    def plot_meshes(self,times=None):  # PLOT ALL THE TIMES STEP EMBRYO IN MORPHONET
        """ Plot all data inside the browser

        Examples
        --------
        >>> mc.plot_meshes()
        """
        if times is None:
            times=range(self.dataset.begin,self.dataset.end+1)
        for t in times:
            self.plot_mesh(t)
           
    def del_mesh(self,t): #DELETE DITECLTY THE OBJ TIME POINT IN UNITY
        """ Delete the specified time point in the browser
        
        Parameters
        ----------
        t : int
            the time point to delete

        Examples
        --------
        >>> mc.del_mesh(1)
        """
        self.send("DEL_"+str(t))

    ################ TO QUIT PROPERLY

    def _clear_temp(self):
        #print(" --> clear temporary path ")
        os.system('rm -rf .temp_morphonet*')

    def _clear_backup(self):
        # print(" --> clear morphonet backup ")
        os.system('rm -rf .backup_morphonet')

    def _receive_signal(self,signalNumber, frame):
        if signalNumber==2:
            try:
                if self.clear_temp:
                    self._clear_temp()
                self.quit()
                quit()
            except:
                print(" --> quit MorphoPlot")
        return






# -*- coding: latin-1 -*-
from morphonet.plugins import MorphoPlugin


class AddTemporalLink(MorphoPlugin):
    """This plugin create a temporal link between objects

    Parameters
    ----------
    Objects: 
        It can be apply either on selected objects or on colored objects where temporal link will done by selection id

    """

    def __init__(self): #PLUGIN DEFINITION 
        MorphoPlugin.__init__(self) 
        self.set_name("Create Links")
        self.set_parent("Temporal Relation")

    def process(self,t,dataset,objects): #PLUGIN EXECUTION
        if not self.start(t,dataset,objects): 
            return None

        #List Objects by selections
        selections={}
        for cid in objects:
            o=dataset.get_object(cid)
            if o is not None:
                if o.s not in selections:
                    selections[o.s]=[]
                selections[o.s].append(o)
        if len(selections)>1:
            self.print_mn(" --> Found  "+str(len(selections))+ " selections ")
        for s in selections:
            #Order objects by time
            times={}  #List all times
            for o in selections[s]:
                if o.t not in times:
                    times[o.t]=[]
                times[o.t].append(o)
            
            for t in sorted(times):
                if t+1 in times:
                    cellT=times[t]
                    cellTP=times[t+1]
                    for daughter in cellTP:
                        for mother in cellT:
                            dataset.add_link(daughter,mother)
                            self.print_mn(" ----> Link object "+str(daughter.id)+" at "+str(daughter.t)+" with "+str(mother.id)+ " at "+str(mother.t))

        self.restart()



# -*- coding: latin-1 -*-
from .MorphoPlugin import MorphoPlugin
__all__ = [
    'MorphoPlugin'
]


defaultPlugins=[]

from .creation import defaultPlugins as DP
defaultPlugins+=DP

from .deletion import defaultPlugins as DP
defaultPlugins+=DP

from .spliting import defaultPlugins as DP
defaultPlugins+=DP

from .temporal import defaultPlugins as DP
defaultPlugins+=DP




def get_borders(data,cellCoords,border=4):
    xmin=max(0,cellCoords[0].min()-border)
    xmax=min(data.shape[0],cellCoords[0].max()+border)
    ymin=max(0,cellCoords[1].min()-border)
    ymax=min(data.shape[1],cellCoords[1].max()+border)
    zmin=max(0,cellCoords[2].min()-border)
    zmax=min(data.shape[2],cellCoords[2].max()+border)
    return xmin,xmax,ymin,ymax,zmin,zmax


def apply_new_label(data,xmin,ymin,zmin,labelw,minVol=0):
    import numpy as np
    labels=np.unique(labelw)
    labels=labels[labels!=0] #Remove Background
    #First We check of all coords have the required minimum of size
    #The biggest cell get the same label
    Labelsize={}
    Bigest=0
    BigestL=-1
    for l in labels:
        Labelsize[l]=len(np.where(labelw==l)[0])
        if Labelsize[l]>Bigest:
            Bigest=Labelsize[l]
            BigestL=l

    labels=labels[labels!=BigestL]
    newIds=[]
    lastID=data.max()+1
    for l in labels:
        new_coords=np.where(labelw==l)
        if len(new_coords[0])>=minVol:
            data[new_coords[0]+xmin,new_coords[1]+ymin,new_coords[2]+zmin]=lastID
            print('     ----->>>>>  Create a new object '+str(lastID)+ " with "+str(len(new_coords[0]))+ " voxels")
            lastID+=1
            newIds.append(lastID)
        else:
            print("     ----->>>>>  Do not create with "+str(len(new_coords[0]))+ " voxels")
    return data,newIds



# -*- coding: latin-1 -*-
defaultPlugins=[]

from .splitInTwoDistancePeak import SplitInTwoDistancePeak
defaultPlugins.append(SplitInTwoDistancePeak())

from .splitOnAxis import SplitOnAxis
defaultPlugins.append(SplitOnAxis())

from .splitOnRaw import SplitOnRaw
defaultPlugins.append(SplitOnRaw())


from .deleteSelectedObjects import *
from .fuseSelectedObjects import *
from .removeUnder import *


# -*- coding: latin-1 -*-
defaultPlugins=[]

from .fuseSelectedObjects import FuseSelectedObjects
defaultPlugins.append(FuseSelectedObjects())

from .deleteSelectedObjects import DeleteSelectedObjects
defaultPlugins.append(DeleteSelectedObjects())


from .removeUnder import RemoveUnder
defaultPlugins.append(RemoveUnder())



# -*- coding: latin-1 -*-
from morphonet.plugins import MorphoPlugin


class Propagate(MorphoPlugin):
    """This plugin propagate a name trough the full cell life

    Parameters
    ----------
    Objects: 
        It can be apply either on selected objects or on colored objects where temporal link will done by selection id

    """

    def __init__(self): #PLUGIN DEFINITION 
        MorphoPlugin.__init__(self) 
        self.set_name("Propagate")
        self.add_inputfield("Info",default="cell_name")
        self.set_parent("Temporal Relation")

    def process(self,t,dataset,objects): #PLUGIN EXECUTION
        if not self.start(t,dataset,objects): 
            return None

        info_name = self.get_inputfield("Info")
        if info_name=="":
            print(" Please fill the info name to propagate")
        else:
            info=dataset.get_info(info_name,create=False)
            if info is None:
                print(" The info named : "+info_name+" is unkown ... ")
            else:
                nb_propagated=0
                for cid in objects:
                    o=dataset.get_object(cid)
                    if o is not None:
                        annotation=info.get(o)
                        if annotation is None:
                            print(" --> no corresponding annotation ")
                        else:
                            print(" --> found Annotation "+annotation)
                            while o.nb_daughters()==1: #Simple propagation method
                                #print(" Cell "+str(o.id)+" at "+str(o.t)+ " is annotated "+str(info.get(o)))
                                if info.get(o)!=annotation:
                                    info.set(o,annotation)
                                    nb_propagated+=1
                                o=o.daughters[0]
                            if info.get(o) != annotation: #The last one
                                info.set(o, annotation)
                                nb_propagated += 1
                print(" --> "+str(nb_propagated)+" info where propagated")
        self.restart() 



# -*- coding: latin-1 -*-
from morphonet.plugins import MorphoPlugin
import operator


class RemoveUnder(MorphoPlugin):
    """ This plugin remove opbjects under a certain volume in the segmented image
  
    Parameters
    ----------
    Voxel Size: int, default 20
        The volume under which objecs as to be remove
    """

    def __init__(self): #PLUGIN DEFINITION 
        MorphoPlugin.__init__(self) 
        self.set_name("Remove Under")
        self.add_inputfield("Voxel Size",default=20)
        self.add_dropdown("Time points",["Current time","All times"])
        self.set_parent("Remove objects")
       

    def process(self,t,dataset,objects): #PLUGIN EXECUTION
        if not self.start(t,dataset,objects,True):
            return None

        import numpy as np
        data = []
        if str(self.get_dropdown("Time points")) == "All times":
            data=dataset.get_all_seg()
            cell_volume = dataset.get_info("cell_volume", info_type="float")
            if data is not None:
                backup_time_list = []
                sorted_info = sorted(cell_volume.data.items(), key=lambda x: x[1][0].value)
                for cell in sorted_info:
                    if cell[1][0].value < float(self.get_inputfield("Voxel Size")) :
                        if dataset.begin <= cell[0].t <= dataset.end:
                            if not cell[0].t in backup_time_list:
                                dataset.backup(self.get_log(), t, objects, True, [cell[0].t,])
                                backup_time_list.append(cell[0].t)
                            coords = np.where(data[t] == cell[0].id)
                            self.print_mn("     ----->>>  delete object " + str(cell[0].id) + " at " + str(cell[0].t) + " with " + str(
                                len(coords[0])) + " pixels")
                            data[cell[0].t][coords] = dataset.background
                            dataset.del_link(cell[0])
                            dataset.set_seg(cell[0].t, data[cell[0].t])
                    else :
                        break
        else :
            data=dataset.get_seg(t)
            cell_volume = dataset.get_info("cell_volume", info_type="float")
            if data is not None:
                backup_time_list = []
                sorted_info = sorted(cell_volume.data.items(), key=lambda x: x[1][0].value)
                for cell in sorted_info:
                    if cell[1][0].value < float(self.get_inputfield("Voxel Size")):
                        if cell[0].t == t:
                            if not cell[0].t in backup_time_list:
                                dataset.backup(self.get_log(), t, objects, True, [t, ])
                                backup_time_list.append(t)
                            coords = np.where(data[t] == cell[0].id)
                            self.print_mn("     ----->>>  delete object " + str(cell[0].id) + " at " + str(
                                cell[0].t) + " with " + str(
                                len(coords[0])) + " pixels")
                            data[cell[0].t][coords] = dataset.background
                            dataset.del_link(cell[0])
                            dataset.set_seg(cell[0].t, data[cell[0].t])
                    else:
                        break
              
        self.restart() #ADD At the end 
         







# -*- coding: latin-1 -*-
import os,sys
import numpy as np
import datetime
from urllib.parse import unquote
from threading import Thread
from os.path import isdir,join,dirname,basename
#****************************************************************** SPECIFC ASTEC FUNCTIONS

#READ AND SAVE IMAGES
def imread(filename):
    """Reads an image file completely into memory

    :Parameters:
     - `filename` (str)

    :Returns Type:
        |numpyarray|
    """
    print(" --> Read "+filename)
    if filename.find('.inr')>0 or filename.find('mha')>0:
        #from morphonet.ImageHandling import SpatialImage
        from morphonet.ImageHandling import imread as imreadINR
        return imreadINR(filename)
    elif filename.find('.nii')>0:
        from nibabel import  load as loadnii
        im_nifti = loadnii(filename)
        return im_nifti
    else:
        from skimage.io import imread as imreadTIFF
        return imreadTIFF(filename)
    return None

def imsave(filename,img):
    """Save a numpyarray as an image to filename.

    The filewriter is choosen according to the file extension. 

    :Parameters:
     - `filename` (str)
     - `img` (|numpyarray|)
    """

    print(" --> Save "+filename)
    if filename.find('.inr')>0 or  filename.find('mha')>0:
        from morphonet.ImageHandling import SpatialImage
        from morphonet.ImageHandling import imsave as imsaveINR
        return imsaveINR(filename,SpatialImage(img))
    elif filename.find('.nii')>0:
        from nibabel import save as savenii
        #new_img = nib.nifti1.Nifti1Image(img, None, header=header_nifti)
        im_nifti = savenii(img,filename)
        return im_nifti

    else:
        from skimage.io import imsave as imsaveTIFF
        return imsaveTIFF(filename,img)
    return None


class _save_seg_thread(Thread):
    #Just perform the saving in thread
    def __init__(self,segment_path,segment_files,data,t,exec_time):
        Thread.__init__(self) 
        self.segment_path=segment_path
        self.segment_files=segment_files
        self.data=data
        self.t=t
        self.exec_time=exec_time

    def run(self): #START FUNCTION
        filename=join(self.segment_path,self.segment_files.format(self.t))
        compressed=False
        if not isfile(filename) and isfile(filename+".gz"):
            compressed=True
        is_save=imsave(filename,self.data)    
        if compressed:
            os.system("gzip -f "+filename)
       

def _add_line_in_file(file,action):
    f = open(file, "a")
    f.write(str(action))
    f.close()
def _read_last_line_in_file(file):
    last_action=""
    for line in open(file, "r"):
        last_action=line
    return last_action

def read_file(filename):
    s=""
    for line in open(filename,"r"):
        s+=line
    return s


def _set_dictionary_value(root):
    """

    :param root:
    :return:
    """

    if len(root) == 0:

        #
        # pas de branche, on renvoie la valeur
        #

        # return ast.literal_eval(root.text)
        if root.text is None:
            return None
        else:
            return eval(root.text)

    else:

        dictionary = {}
        for child in root:
            key = child.tag
            if child.tag == 'cell':
                key = np.int64(child.attrib['cell-id'])
            dictionary[key] = _set_dictionary_value(child)

    return dictionary


#Read XML Properties
def read_XML_properties(filename):
    """
    Return a xml properties from a file 
    :param filename:
    :return as a dictionnary
    """
    properties = None
    if not os.path.exists(filename):
        print(' --> properties file missing '+filename)
    elif filename.endswith("xml") is True:
        print(' --> read XML properties from '+filename)
        import xml.etree.ElementTree as ElementTree
        inputxmltree = ElementTree.parse(filename)
        root = inputxmltree.getroot()
        properties= _set_dictionary_value(root)
    else:
        print(' --> unkown properties format for '+filename)
    return properties


def _indent(elem, level=0):
    i = "\n" + level*"  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            _indent(elem, level+1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i



def _set_xml_element_text(element, value):
    """

    :param element:
    :param value:
    :return:
    """
    #
    # dictionary : recursive call
    #   dictionary element may be list, int, numpy.ndarray, str
    # list : may be int, numpy.int64, numpy.float64, numpy.ndarray
    #

    if type(value) == dict:
        # print proc + ": type is dict"
        keylist = value.keys()
        sorted(keylist)
        for k in keylist:
            _dict2xml(element, k, value[k])

    elif type(value) == list:

        #
        # empty list
        #

        if len(value) == 0:
            element.text = repr(value)
        #
        # 'lineage', 'label_in_time', 'all-cells', 'principal-value'
        #

        elif type(value[0]) in (int, float, np.int64, np.float64):
            # element.text = str(value)
            element.text = repr(value)

        #
        # 'principal-vector' case
        #  liste de numpy.ndarray de numpy.float64
        #
        elif type(value[0]) == np.ndarray:
            text = "["
            for i in range(len(value)):
                # text += str(list(value[i]))
                text += repr(list(value[i]))
                if i < len(value)-1:
                    text += ", "
                    if i > 0 and i % 10 == 0:
                        text += "\n  "
            text += "]"
            element.text = text
            del text

        else:
            element.text = repr(value)
            #print( " --> error, element list type ('" + str(type(value[0]))  + "') not handled yet for "+str(value))
            #quit()
    #
    # 'barycenter', 'cell_history'
    #
    elif type(value) == np.ndarray:
        # element.text = str(list(value))
        element.text = repr(list(value))

    #
    # 'volume', 'contact'
    #
    elif type(value) in (int, float, np.int64, np.float64):
        # element.text = str(value)
        element.text = repr(value)

    #
    # 'fate', 'name'
    #
    elif type(value) == str:
        element.text = repr(value)

    else:
        print( " --> element type '" + str(type(value))  + "' not handled yet, uncomplete translation")
        quit()

def _dict2xml(parent, tag, value):
    """

    :param parent:
    :param tag:
    :param value:
    :return:
    """

    #
    # integers can not be XML tags
    #
    import xml.etree.ElementTree as ElementTree
    if type(tag) in (int, np.int64):
        child = ElementTree.Element('cell', attrib={'cell-id': str(tag)})
    else:
        child = ElementTree.Element(str(tag))

    _set_xml_element_text(child, value)
    parent.append(child)
    return parent

def dict2xml(dictionary, defaultroottag='data'):
    """

    :param dictionary:
    :param defaultroottag:
    :return:
    """
    import xml.etree.ElementTree as ElementTree
    if type(dictionary) is not dict:
        print(" --> error, input is of type '" + str(type(dictionary)) + "'")
        return None

    if len(dictionary) == 1:
        roottag = dictionary.keys()[0]
        root = ElementTree.Element(roottag)
        _set_xml_element_text(root, dictionary[roottag])

    elif len(dictionary) > 1:
        root = ElementTree.Element(defaultroottag)
        for k, v in dictionary.items():
            _dict2xml(root, k, v)

    else:
        print(" --> error, empty dictionary ?!")
        return None

    _indent(root)
    tree = ElementTree.ElementTree(root)

    return tree

def write_XML_properties(properties,filename,thread_mode=True):
    """
    Write a xml properties in a file 
    :param properties:
    :param filename:
    """
    if thread_mode:
        wxml = Thread(target=write_XML_properties_thread, args=[properties,filename])
        wxml.start()
    else:
        write_XML_properties_thread(properties,filename)

def write_XML_properties_thread(properties,filename):
    """
    Write a xml properties in a file in Thread Mode
    :param properties:
    :param filename:
    """
    if properties is not None:
        xmltree=dict2xml(properties)
        print(" --> write XML properties in "+filename)
        xmltree.write(filename)




#Return t, cell_id from long name : t*10**4+id (to have an unique identifier of cells)
def get_id_t(idl):
    t=int(int(idl)/(10**4))
    cell_id=int(idl)-int(t)*10**4
    return t,cell_id
def get_longid(t,idc):
    return t*10**4+idc
 

#Return Cell name as string
def get_name(t,id):
    return str(t)+","+str(id)

def _get_object(o):
    """ Construct an object (as a tuple) from a string
        
    """
    to=0
    ido=0
    cho=0
    oss=o.split(',')
    if len(oss)==1:
        ido=int(o)
    if len(oss)>1:
        to=int(oss[0])
        ido=int(oss[1])
    if len(oss)>2:
        cho=int(oss[2])
    if cho==0:
        return (to, ido) #We do not put channel 0 for most of the case
    return (to,ido,cho)



def _get_objects(infos):
        """ Get the list of object from an information data
        
        Parameters
        ----------
        infos : string
            The information data

        Returns
        -------
        objects : list
            List of key/value corresponding to a split of the data

        """
        if type(infos)==bytes or type(infos)==bytearray:
            infos=infos.decode('utf-8')
        #print(type(infos))
        infos=infos.split("\n")
        objects={}
        for line in infos:
            if len(line)>0 and line[0]!="#":
                if line.find("type")==0:
                    dtype=line.replace("type:","")
                else:
                    tab=line.split(":")
                    ob=_get_object(tab[0])
                    if ob in objects: #Multiple times the same value (we put in list)
                        val1=objects[ob]
                        if type(val1)!=list :
                            objects[ob]=[]
                            objects[ob].append(val1)
                        if dtype =="time" or dtype =="space" :
                            objects[ob].append(_get_object(tab[1]))
                        elif dtype == "dict":
                            objects[ob].append((_get_object(tab[1]), tab[2]))
                        else:
                            objects[ob].append(tab[1])
                    else:
                        if dtype =="time" or dtype =="space" :
                            objects[ob]=_get_object(tab[1])
                        elif dtype=="dict": #178,724:178,1,0:602.649597
                            objects[ob]=[]
                            objects[ob].append((_get_object(tab[1]),tab[2]))
                        else:
                            objects[ob] = tab[1]

        return objects



def _get_type(infos):
        """ Get the type from an information data
        
        Parameters
        ----------
        infos : string
            The information data

        Returns
        -------
        type : string
            the type (float, string, ...)

        """
        infos=infos.split('\n')
        for line in infos:
            if len(line)>0 and line[0]!="#":
                if line.find("type")==0:
                    return line.split(":")[1]
        return None

def _get_string(ob):
    return str(ob[0])+","+str(ob[1])+","+str(ob[2])

def _get_last_curation(l):
    if type(l)==list:
        lastD=datetime.datetime.strptime('1018-06-29 08:15:27','%Y-%m-%d %H:%M:%S')
        value=""
        for o in l:
            d=o.split(";")[2] #1 Value, 2 Guy, 3 Date
            d2 = datetime.datetime.strptime(d,'%Y-%m-%d-%H-%M-%S')
            if d2>lastD:
                lastD=d2
                value=o
        return value
    return l


def _get_param(command,p): #Return the value of a specific parameter in http query
    params=unquote(str(command.decode('utf-8'))).split("&")
    for par in params:
        k=par.split("=")[0]
        if k==p:
            return par.split("=")[1].replace('%20',' ')
    return ""



def isfile(filename):
    if os.path.isfile(filename):
        return True
    elif os.path.isfile(filename+".gz"):
        return True
    elif os.path.isfile(filename+".zip"):
        return True
    return False

def copy(filename1,filname2):
    if os.path.isfile(filename1):
        os.system('cp '+filename1+" "+filname2)
    elif os.path.isfile(filename1+".gz"):
        os.system('cp '+filename1+".gz "+filname2+".gz")
    elif os.path.isfile(filename1+".zip"):
        os.system('cp '+filename1+".zip "+filname2+".zip")
    else:
        print("ERROR didn't found to copy "+filename1)

def load_mesh(filename,voxel_size=None,center=None):
    f=open(filename,'r')
    obj=''
    for line in f:
        if len(line)>4 and line.find("v")==0 and line[1]==" ": #VERTEX
            if voxel_size is not None or center is not None:
                tab=line.replace('\t',' ').replace('   ',' ').replace('  ',' ').split(" ")
                v=[float(tab[1]),float(tab[2]),float(tab[3])]
                if voxel_size is not None:
                    if type(voxel_size)==str:
                        vs=voxel_size.split(",")
                        if len(vs)==3:
                            v[0]= v[0]*float(vs[0])
                            v[1]= v[1]*float(vs[1])
                            v[2]= v[2]*float(vs[2])
                    else:
                        v=v*voxel_size
                if center is not None:
                    v=v-center
                obj+="v "+str(v[0])+" "+str(v[1])+" "+str(v[2])+"\n"
            else:
                obj += line
        else:
            obj+=line
    f.close()
    return obj

def save_mesh(filename,obj):
    f = open(filename, "w")
    f.write(obj)
    f.close()

def get_objects_by_time(dataset,objects):
    times=[]
    for cid in objects: #List all time points
        o=dataset.get_object(cid)
        if o is not None and o.t not in times:
            times.append(o.t)
    times.sort() #Order Times
    return times

from vtk import vtkImageImport,vtkDiscreteMarchingCubes,vtkWindowedSincPolyDataFilter,vtkQuadricClustering,vtkDecimatePro,vtkPolyDataReader,vtkPolyDataWriter    
from threading import Thread
_dataToConvert=None
class convert_one_to_OBJ(Thread):
    def __init__(self, t,elt,Smooth,Decimate,Reduction,path_write,recompute,TargetReduction=0.8):
        Thread.__init__(self)
        self.t=t
        self.elt = elt
        self.Smooth = Smooth
        self.Decimate = Decimate
        self.Reduction = Reduction
        self.TargetReduction=TargetReduction
        self.polydata=None
        self.recompute=True
        self.filename=None
        if path_write is not None:
            self.recompute=recompute
            self.filename=os.path.join(path_write,str(t)+'-'+str(elt)+'.vtk')
    def run(self):
        global _dataToConvert
        if not self.recompute:
            self.recompute=self.read()
        if self.recompute:
            #print(" Compute "+str(self.t) +"-"+str(self.elt))
            nx, ny, nz = _dataToConvert.shape
            eltsd=np.zeros(_dataToConvert.shape,np.uint8)
            coord=np.where(_dataToConvert==self.elt)
            #print('     ----->>>>> Create cell '+str(self.elt) + " with "+str(len(coord[0]))+' pixels ')
            eltsd[coord]=255

            data_string = eltsd.tostring('F')
            reader = vtkImageImport()
            reader.CopyImportVoidPointer(data_string, len(data_string))
            reader.SetDataScalarTypeToUnsignedChar()

            reader.SetNumberOfScalarComponents(1)
            reader.SetDataExtent(0, nx - 1, 0, ny - 1, 0, nz - 1)
            reader.SetWholeExtent(0, nx - 1, 0, ny - 1, 0, nz - 1)
            reader.Update()

            #MARCHING CUBES
            contour = vtkDiscreteMarchingCubes()
            contour.SetInputData(reader.GetOutput())
            contour.ComputeNormalsOn()
            contour.ComputeGradientsOn()
            contour.SetValue(0,255)
            contour.Update()
            self.polydata= contour.GetOutput()

            if self.Smooth and self.polydata.GetPoints() is not None:
                smooth_angle=120.0
                smoth_passband=0.01
                smooth_itertations=25
                smoother = vtkWindowedSincPolyDataFilter()
                smoother.SetInputData(self.polydata)
                smoother.SetFeatureAngle(smooth_angle)
                smoother.SetPassBand(smoth_passband)
                smoother.SetNumberOfIterations(smooth_itertations)
                smoother.NonManifoldSmoothingOn()
                smoother.NormalizeCoordinatesOn()
                smoother.Update()
                self.polydata= smoother.GetOutput()


            if self.Decimate and self.polydata.GetPoints() is not None:
                mesh_fineness=1.0
                decimater = vtkQuadricClustering()
                decimater.SetInputData(self.polydata)
                decimater.SetNumberOfDivisions(*np.uint16(tuple(mesh_fineness*np.array(np.array(_dataToConvert.shape)/2))))
                decimater.SetFeaturePointsAngle(30.0)
                decimater.CopyCellDataOn()
                decimater.Update()
                self.polydata= decimater.GetOutput()

            if self.Reduction and self.polydata.GetPoints() is not None:
                decimatePro  = vtkDecimatePro()
                decimatePro.SetInputData(self.polydata)
                decimatePro.SetTargetReduction(self.TargetReduction)
                decimatePro.Update()
                self.polydata= decimatePro.GetOutput()
    
    def read(self):
        if os.path.isfile(self.filename):
            #print("Read "+self.filename)
            reader = vtkPolyDataReader()
            reader.SetFileName(self.filename)
            reader.Update()
            self.polydata=reader.GetOutput()
            return False
        return True


    def write(self):
        if self.recompute and self.filename is not None:
            #print("Write "+self.filename)
            writer = vtkPolyDataWriter()
            writer.SetFileName(self.filename)
            writer.SetInputData(self.polydata)
            writer.Update()
 

def convert_to_OBJ(dataFull,t,background=0,factor=1,Smooth=True,Decimate=True,Reduction=True,TargetReduction=0.8,Border=2,center=[0,0,0],VoxelSize=[1,1,1],maxNumberOfThreads=None,cells_updated=None,path_write=None): ####  CONVERT SEGMENTATION IN MESH
        global _dataToConvert
        if maxNumberOfThreads is None:
            maxNumberOfThreads=os.cpu_count()*2
        _dataToConvert=dataFull[::factor,::factor,::factor]
        if Border>0: #We add border to close the cell
            _dataToConvert=np.zeros(np.array(_dataToConvert.shape) + Border * 2).astype(dataFull.dtype)
            _dataToConvert[:,:,:]=background
            _dataToConvert[Border:-Border,Border:-Border,Border:-Border]=dataFull[::factor,::factor,::factor]
        elts=np.unique(_dataToConvert)
        elts=elts[elts!=background] #Remove Background

        threads=[]
        recompute=path_write is not None and cells_updated is not None
        all_threads=[]
        for elt in elts:
            if len(threads)>=maxNumberOfThreads:
                tc = threads.pop(0)
                tc.join()
                tc.write()

            #print(" Compute cell "+str(elt))
            recompute_cell=True if cells_updated is None else elt in cells_updated
            tc = convert_one_to_OBJ(t, elt, Smooth, Decimate, Reduction, path_write, recompute_cell,TargetReduction=TargetReduction)
            tc.start()
            all_threads.append(tc)
            threads.append(tc)

        #Finish all threads left
        while len(threads)>0:
            tc = threads.pop(0)
            tc.join()
            tc.write()

        #Merge all polydata in one
        obj=""
        shiftFace=1
        for tc in all_threads:
            polydata=tc.polydata
            elt=tc.elt
            if polydata.GetPoints() is not None:
                obj+="g "+str(t)+","+str(elt)+"\n"
                if not polydata.GetPoints() is None :
                    for p in range(polydata.GetPoints().GetNumberOfPoints()):
                        v=polydata.GetPoints().GetPoint(p)
                        obj+='v ' + str((v[0]-Border)*factor*VoxelSize[0]-center[0]) +' '+str((v[1]-Border)*factor*VoxelSize[1]-center[1]) +' '+str((v[2]-Border)*factor*VoxelSize[2]-center[2])+'\n'
                    for f in range(polydata.GetNumberOfCells()):
                        obj+='f ' + str(shiftFace+polydata.GetCell(f).GetPointIds().GetId(0)) +' '+str(shiftFace+polydata.GetCell(f).GetPointIds().GetId(1)) +' '+str(shiftFace+polydata.GetCell(f).GetPointIds().GetId(2))+'\n'
                    shiftFace+=polydata.GetPoints().GetNumberOfPoints()
        return obj


def add_slashes(s):
    d = {'"':'\\"', "'":"\\'", "\0":"\\\0", "\\":"\\\\"}
    return ''.join(d.get(c, c) for c in s)


def try_parse_int(value):
    if value is None:
        return None
    try:
        return int(value)
    except ValueError:
        return None
    return None

class bcolors:
    HEADER = '\033[95m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

ss=" --> "
def strblue(strs):
    return bcolors.BLUE+strs+bcolors.ENDC
def strred(strs):
    return bcolors.RED+strs+bcolors.ENDC
def strgreen(strs):
    return bcolors.BOLD+strs+bcolors.ENDC

def nodata(data,other_test=None):
    if data=="" or data==[] or len(data)==0:
        return True
    if type(data)==str:
        if data.lower().find("done")>=0 or data.lower().find("status")>=0:
            return True
    if type(data)==dict:
        if "status" in data and data['status'].lower()=="error":
            return True
    if other_test is not None:
        if other_test not in data:
            return True
    return False

def error_request(data,msg):
    if "error_message" in data:
        print(strred(" --> Error "+msg+" : "+data["error_message"]))
    else:
        print(strred(" --> Error "+msg +" : with not error message"))
    return False

def _get_pip_version(projet="morphonet"):
    '''
    Find the last available version of MorphoNet API
    '''
    import urllib.request
    fp = urllib.request.urlopen("https://pypi.org/project/"+projet)
    release__version=False
    for lines in fp.readlines():
        if release__version:
            return lines.decode("utf8").strip()
        if lines.decode("utf8").find("release__version")>0:
            release__version=True
    return "unknown"

def _check_version():
    '''
    Chekc if the API installed is the last version
    '''
    import pkg_resources
    current_version=None
    try :
        current_version = pkg_resources.get_distribution('morphonet').version
    except:
        print(' --> did not find current version of MorphoNet API ')

    online_version=None
    try:
        online_version = _get_pip_version()
    except:
        print(' --> did not last version of MorphoNet API ')

    if current_version is not None and online_version is not None and current_version != online_version:
        print(strblue("WARNING : please update your MorphoNet version : pip install -U morphonet "))
        return False
    return True


def get_fate_colormap(fate_version):

    ColorFate2020 = {}
    ColorFate2020["1st Lineage, Notochord"] = 2
    ColorFate2020["Posterior Ventral Neural Plate"] = 19
    ColorFate2020["Anterior Ventral Neural Plate"] = 9
    ColorFate2020["Anterior Head Endoderm"] = 8
    ColorFate2020["Anterior Endoderm"] = 8
    ColorFate2020["Posterior Head Endoderm"] = 17
    ColorFate2020["Posterior Endoderm"] = 17
    ColorFate2020["Trunk Lateral Cell"] = 20
    ColorFate2020["Mesenchyme"] = 14
    ColorFate2020["1st Lineage, Tail Muscle"] = 3
    ColorFate2020["Trunk Ventral Cell"] = 21
    ColorFate2020["Germ Line"] = 10
    ColorFate2020["Lateral Tail Epidermis"] = 12
    ColorFate2020["Head Epidermis"] = 11
    ColorFate2020["Trunk Epidermis"] = 11
    ColorFate2020["Anterior Dorsal Neural Plate"] = 7
    ColorFate2020["Posterior Lateral Neural Plate"] = 18
    ColorFate2020["2nd Lineage, Notochord"] = 5
    ColorFate2020["Medio-Lateral Tail Epidermis"] = 13
    ColorFate2020["Midline Tail Epidermis"] = 15
    ColorFate2020["Posterior Dorsal Neural Plate"] = 16
    ColorFate2020["1st Endodermal Lineage"] = 1
    ColorFate2020["2nd Lineage, Tail Muscle"] = 6
    ColorFate2020["2nd Endodermal Lineage"] = 4

    ColorFate2009 = {}
    ColorFate2009["1st Lineage, Notochord"] = 78
    ColorFate2009["Posterior Ventral Neural Plate"] = 58
    ColorFate2009["Anterior Ventral Neural Plate"] = 123
    ColorFate2009["Anterior Head Endoderm"] = 1
    ColorFate2009["Anterior Endoderm"] = 1
    ColorFate2009["Posterior Head Endoderm"] = 27
    ColorFate2009["Posterior Endoderm"] = 27
    ColorFate2009["Trunk Lateral Cell"] = 62
    ColorFate2009["Mesenchyme"] = 63
    ColorFate2009["1st Lineage, Tail Muscle"] = 135
    ColorFate2009["Trunk Ventral Cell"] = 72
    ColorFate2009["Germ Line"] = 99
    ColorFate2009["Lateral Tail Epidermis"] = 61
    ColorFate2009["Head Epidermis"] = 76
    ColorFate2020["Trunk Epidermis"] = 76
    ColorFate2009["Anterior Dorsal Neural Plate"] = 81
    ColorFate2009["Posterior Lateral Neural Plate"] = 75
    ColorFate2009["2nd Lineage, Notochord"] = 199
    ColorFate2009["Medio-Lateral Tail Epidermis"] = 41
    ColorFate2009["Midline Tail Epidermis"] = 86
    ColorFate2009["Posterior Dorsal Neural Plate"] = 241
    ColorFate2009["1st Endodermal Lineage"] = 40
    ColorFate2009["2nd Lineage, Tail Muscle"] = 110
    ColorFate2009["2nd Endodermal Lineage"] = 44

    if fate_version=="2020":
        return ColorFate2020
    return ColorFate2009

def get_info_from_properties(prop,info_name,info_type,convert=None):
    info="#" + info_name+"\n"
    if type(prop) == list:
        info_type="selection"
    info+= "type:" + info_type + "\n"
    Missing_Conversion=[]
    if type(prop) == list:
        for idl in prop:
            t, c = get_id_t(idl)
            info += get_name(t, c) + ":1\n"
    else:
        for idl in prop:
            t, c = get_id_t(idl)
            if info_type == 'time':
                for daughter in prop[idl]:
                    td, d = get_id_t(daughter)
                    info+=get_name(t, c) + ":" + get_name(td, d) + "\n"
            elif info_type == 'dict':  #178,724:178,1,0:602.649597
                for elt in prop[idl]:
                    td, d = get_id_t(elt)
                    info += get_name(t, c) + ":" + get_name(td, d)+":"+str(prop[idl][elt]) + "\n"
            else:
                if convert is None:
                    if type(prop[idl])== list:
                        for elt in prop[idl]:
                            info+=get_name(t, c) + ":" + str(elt) + "\n"
                    else:
                        info+=get_name(t, c) + ":" + str(prop[idl]) + "\n"
                else:
                    if type(prop[idl])== list:
                        for elt in prop[idl]:
                            if elt not in convert:
                                if elt not in Missing_Conversion:
                                    Missing_Conversion.append(elt)
                            else:
                                info += get_name(t, c) + ":" + str(convert[elt]) + "\n"
                    else:
                        if prop[idl] not in convert:
                            if prop[idl] not in Missing_Conversion:
                                Missing_Conversion.append(prop[idl])
                        else:
                            info += get_name(t, c) + ":" + str(convert[prop[idl]]) + "\n"
    for elt in Missing_Conversion:
        print(" ->> Misss '" + str(elt) + "' in the selection conversion ")
    return info

def write_info(filename,prop,info_name,info_type,convert=None):
    if info_type is None:
        info_type = get_info_type(info_name)
    if info_type is None:
        print(" ->> Did not find type for " + info_name)
    else:
        print(" Write "+filename)
        f = open(filename, "w")
        f.write(get_info_from_properties(prop,info_name.replace("selection_", ""),info_type,convert=convert))
        f.close()

def get_info_type(info_name):
    if info_name.lower().startswith("selection"):
        return "selection"
    if info_name.lower().find("lineage") >= 0:
        return "time"
    if info_name.lower().find("cell_contact_surface") >= 0:
        return "dict"
    if info_name.lower().find("surface") >= 0:
        return "float"
    if info_name.lower().find("volume") >= 0:
        return "float"
    if info_name.lower().find("area") >= 0:
        return "float"
    if info_name.lower().find("fate") >= 0:
        return "string"
    if info_name.lower().find("name") >= 0:
        return "string"
    return None



# -*- coding: latin-1 -*-
from morphonet.plugins import MorphoPlugin

class CreateSeeds(MorphoPlugin):
    """ This plugin create new seeds from a h min (or max) algorithm on the rawdata image

    Parameters
    ----------
    Gaussian_Sigma : int, default :8
        sigma parameters from the gaussian algorithm (from skimage) applied on the rawdata in otder to perform the h minimum or maximum algorithm
    MinOrMax: Dropdown
        To aplly the h minmum or h maximum algorithm (depend on the color of the bacground of the rawdata )
    h_value : int, default :2
        the h value of h_minima or h_maxumum algorithm (see https://scikit-image.org/docs/stable/api/skimage.morphology.html )
        
    """

    def __init__(self): #PLUGIN DEFINITION 
        MorphoPlugin.__init__(self) 
        self.set_name("CreateSeeds")
        self.add_inputfield("gaussian_sigma",default=8)
        self.add_dropdown("MinOrMax",["min","max"])
        self.add_inputfield("h_value",default=2)
        self.set_parent("Create new objects")

    def process(self,t,dataset,objects): #PLUGIN EXECUTION
        if not self.start(t,dataset,objects):
            return None
  
        s_sigma=int(self.get_inputfield("gaussian_sigma"))
        MinOrMax=self.get_dropdown("MinOrMax")
        h_value=float(self.get_inputfield("h_value"))
       
        from skimage.morphology import extrema
        from skimage.filters import gaussian 
        from skimage.measure import label
        import numpy as np

        data=dataset.get_seg(t)
        center=dataset.get_center(data)
        rawdata=dataset.get_raw(t)

        #Smoothing
        if s_sigma > 0.0:
            self.print_mn(" --> Perform gaussian with sigma="+str(s_sigma))
            seed_preimage=gaussian(rawdata, sigma=s_sigma,preserve_range=True)
        else:
            seed_preimage = rawdata

        #Fin Extrema
        self.print_mn(" --> Perform  h "+MinOrMax+ " whith h = "+str(h_value))
        if MinOrMax=="min":
            local=extrema.h_minima(seed_preimage,h_value)
        else: 
            local=extrema.h_maxima(seed_preimage,h_value) 

        self.print_mn(" --> Perform  labelisation")
        label_maxima,nbElts = label(local,return_num=True) 
        nbc=0
        for elt in range(1,nbElts+1):
            coord=np.where(label_maxima==elt)
            v=dataset.background
            if data is not None:
                v=data[coord[0][0],coord[1][0],coord[2][0]]
            if v==dataset.background:
                dataset.add_seed(np.int32([coord[0][0]-center[0],coord[1][0]-center[1],coord[2][0]-center[2]]))
                self.print_mn(" ----> Create a seed at "+str(coord[0][0])+","+str(coord[1][0])+","+str(coord[2][0]))
                nbc+=1
        self.print_mn(" --> Found "+str(nbc)+" peaks with "+MinOrMax+" extrema")

        self.restart()




# -*- coding: latin-1 -*-
from morphonet.plugins import MorphoPlugin
from .functions import *

class SplitInTwoDistancePeak(MorphoPlugin):
    """ This plugin split opbjects based on distance between two Peaks in the segmented image 

    Parameters
    ----------
    Objects: 
        It can be apply either on selected or colored objects

    """
    
    def __init__(self): #PLUGIN DEFINITION 
        MorphoPlugin.__init__(self) 
        self.set_name("Split In 2")
        self.set_parent("Split objects")

    def process(self,t,dataset,objects): #PLUGIN EXECUTION
        if not self.start(t,dataset,objects):
            return None
            
        from scipy import ndimage as ndi 
        from skimage.segmentation import watershed
        from skimage.feature import peak_local_max
        import numpy as np
        for cid in objects:
            o=dataset.get_object(cid)
            if o is not None:
                data=dataset.get_seg(o.t)
                cellCoords=np.where(data==o.id)
                self.print_mn('     ----->>>  Split object '+str(o.get_name()) + " with "+str(len(cellCoords[0]))+ " voxels ")
                xmin,xmax,ymin,ymax,zmin,zmax=get_borders(data,cellCoords)
                cellShape=[1+xmax-xmin,1+ymax-ymin,1+zmax-zmin]
                mask=np.zeros(cellShape,dtype=np.bool)
                mask[cellCoords[0]-xmin,cellCoords[1]-ymin,cellCoords[2]-zmin]=True
                distance = ndi.distance_transform_edt(mask)
                peak_idx = peak_local_max(distance,  footprint=np.ones((3, 3, 3)),num_peaks=2)
                local_maxi = np.zeros_like(mask, dtype=bool)
                local_maxi[tuple(peak_idx.T)] = True
                markers = ndi.label(local_maxi)[0]
                labelw = watershed(-distance, markers, mask=mask)
                data,newIds=apply_new_label(data,xmin,ymin,zmin,labelw)
                if len(newIds)>0:
                    dataset.del_link(o)
                    newIds.append(o.id)
                    dataset.set_seg(t,data,newIds)
                
        self.restart()



# -*- coding: latin-1 -*-
from morphonet.plugins import MorphoPlugin



class SplitOnAxis(MorphoPlugin):
    """ This plugin split objects in two on a given axis

    Parameters
    ----------
    Objects: 
        It can be apply either on selected or colored objects
    Axis : Dropdown
        It can be X,Y or Z axis

    """
    def __init__(self): #PLUGIN DEFINITION 
        MorphoPlugin.__init__(self) 
        self.set_name("Split On Axis")
        self.add_dropdown("Axis",["X","Y","Z"])
        self.set_parent("Split objects")

    def process(self,t,dataset,objects): #PLUGIN EXECUTION
        if not self.start(t,dataset,objects):
            return None

        import numpy as np
        which=self.get_dropdown("Axis")
        xyz=-1
        if which=="X":
            xyz=0
        elif which=="Y":
            xyz=1
        elif which=="Z":
            xyz=2
        if xyz==-1:
            self.print_mn('ERROR' + which+ " unknown ....")
        else:
            for cid in objects:
                o=dataset.get_object(cid)
                if o is not None:
                    self.print_mn('     ----->>>  Split Object '+str(o.get_name())+ " in "+str(which))
                    data=dataset.get_seg(o.t)
                    coords=np.where(data==o.id)
                    xyzList=np.unique(coords[xyz])
                    xyzList.sort()
                    lastID=data.max()
                    lastID=lastID+1
                    w=np.where(coords[xyz]>int(xyzList.mean()))
                    new_coords=(coords[0][w],coords[1][w],coords[2][w])
                    data[new_coords]=lastID
                    self.print_mn('     ----->>>>>  Create a new ID '+str(lastID)+ " with "+str(len(new_coords[0]))+ " pixels")
                    dataset.set_seg(t,data,[o.id,lastID])

        self.restart()



# -*- coding: latin-1 -*-
from morphonet.plugins import MorphoPlugin


# Test if a center is inside the image
def _centerInShape(c,s): 
    if c[0]<0 or c[1]<0 or c[2]<0 or c[0]>=s[0] or c[1]>=s[1] or c[2]>=s[2]:
        return False 
    return True


class WatershedithNewSeeds(MorphoPlugin):
    """ This plugin perform a watershed algorithm on the background of the image based on seeds pass in parameters
   
    Parameters
    ----------
    Gaussian_Sigma : int, default :2
        sigma parameters from the gaussian algorithm (from skimage) aplied on the rawdata
    Volume_Minimum: int, default : 1000
        minimum volume under wichi new object are created
    Inverse: Dropdown
        applied the watershed on inverted rawdata (for image on black or white background)
    Seeds: Coordinate List
        List of seeds added on the MorphoNet Window

    """

    def __init__(self): #PLUGIN DEFINITION 
        MorphoPlugin.__init__(self) 
        self.set_name("Watershed")
        self.add_inputfield("Gaussian_Sigma",default=2)
        self.add_inputfield("Volume_Minimum",default=1000)
        self.add_dropdown("Inverse",["no","yes"])
        self.add_coordinates("Add a Seed")
        self.set_parent("Create new objects")

    def process(self,t,dataset,objects): #PLUGIN EXECUTION
        if not self.start(t,dataset,objects):
            return None
        seeds=self.get_coordinates("Add a Seed")
        if len(seeds)==0:
            self.restart()
            return None

        s_sigma=int(self.get_inputfield("gaussian_sigma"))
        Inverse=self.get_dropdown("Inverse")
        min_vol=int(self.get_inputfield("Volume_Minimum"))

        from skimage.morphology import label
        from skimage.segmentation import watershed
        from skimage.filters import gaussian 
        import numpy as np
        data=dataset.get_seg(t)
        rawdata=dataset.get_raw(t)

        if data is None:
            data=np.zeros(rawdata.shape).astype(np.uint16)

        center=dataset.get_center(data)
        if center is None:
            center=[np.round(rawdata.shape[0]/2),np.round(rawdata.shape[1]/2),np.round(rawdata.shape[2]/2)]
        #First we remove seeds which are inside other cell 

        #Smoothing
        if s_sigma > 0.0:
            self.print_mn(" --> Perform gaussian with sigma="+str(s_sigma))
            seed_preimage=gaussian(rawdata, sigma=s_sigma,preserve_range=True)
        else:
            seed_preimage = rawdata

        new_seed=[]
        
        for s in seeds:
            seed=np.int32(s+center)
            if _centerInShape(seed,data.shape):
                olid=data[seed[0],seed[1],seed[2]]
                if olid==dataset.background: 
                    new_seed.append(seed)
                    self.print_mn(" ----> add seed "+str(seed))
                else:
                    self.print_mn(" ----> remove this seed "+str(seed)+ " which already correspond to cell "+str(olid))
            else:
                self.print_mn(" ----> this seed "+str(seed)+ " is out of the image")

                if len(new_seed)==0:
                    self.restart()
                    return None


        markers=np.zeros(data.shape,dtype=np.uint16)
        markers[0,:,:]=1
        markers[:,0,:]=1
        markers[:,:,0]=1
        markers[data.shape[0]-1,:,:]=1
        markers[:,data.shape[1]-1,:]=1
        markers[:,:,data.shape[2]-1]=1
        
        newId=2
        for seed in new_seed: #For Each Seeds ...
            markers[seed[0],seed[1],seed[2]]=newId
            newId+=1
                
        #Create The Mask
        mask=np.ones(data.shape,dtype=np.bool)
        mask[data!=dataset.background]=False
            
        #Inverse the image
        if Inverse=="yes":
            seed_preimage=seed_preimage.max()-seed_preimage

        self.print_mn(" --> Process watershed ")
        labelw=watershed(seed_preimage,markers=markers, mask=mask)

        self.print_mn(" --> Combine new objects")
        cMax=data.max()
        new_ids=np.unique(labelw)
        new_ids=new_ids[new_ids>1]
        nbc=1

        for new_id in new_ids:
            newIdCoord=np.where(labelw==new_id)
            if len(newIdCoord[0])>min_vol:
                data[newIdCoord]=cMax+nbc
                self.print_mn(" ----> add object "+str(nbc+cMax)+' with  '+str(len(newIdCoord[0]))+ " voxels")
                dataset.set_seg(t,data,[new_id])
                nbc+=1
            else:
                self.print_mn(" ----> remove object with  "+str(len(newIdCoord[0]))+ " voxels")
        self.print_mn(" --> Found  "+str(nbc)+" new labels")
        self.restart()







#!/usr/bin/env python3
#
#       image.serial: read/write spatial nd images
#
#       Copyright 2006 - 2018 INRIA - CIRAD - INRA
#
#       File author(s): Jerome Chopard <jerome.chopard@sophia.inria.fr>
#                       Eric Moscardi <eric.moscardi@sophia.inria.fr>
#                       Daniel Barbeau <daniel.barbeau@sophia.inria.fr>
#                       Gregoire Malandain
#
#       Distributed under the Cecill-C License.
#       See accompanying file LICENSE.txt or copy at
#           http://www.cecill.info/licences/Licence_CeCILL-C_V1-en.html
#
#       OpenAlea WebSite: http://openalea.gforge.inria.fr
################################################################################
"""
This module defines inrimage format
"""

__license__ = "Cecill-C"
__revision__ = " $Id$ "

#import os
import os
from os import path
import sys
import numpy as np
# from struct import calcsize, pack, unpack
import gzip
#import io
from io import BytesIO
from .spatial_image import SpatialImage

__all__ = ["read_metaimage", "write_metaimage"]

specific_header_keys = ("ObjectType",
                        "NDims", "DimSize", "ElementNumberOfChannels",
                        "ElementSize", "ElementSpacing", "ElementType",
                        "CompressedData", "BinaryDataByteOrderMSB", "BinaryData",
                        "ElementDataFile")


def open_metaimagefile(filename):
    """Open an inrimage file

    Manage the gz attribute
    """
    program = "open_inrifile"
    if not os.path.isfile(filename) and os.path.isfile(filename+".gz"):
        filename = filename+".gz"
        #print ("%s: Warning: path to read image has been changed to %s." % (program, filename))
    if not os.path.isfile(filename) and os.path.isfile(filename+".zip"):
        filename = filename+".zip"
        #print ("%s: Warning: path to read image has been changed to %s." % (program, filename))
    if path.splitext(filename)[1] in (".gz", ".zip"):
        fzip = gzip.open(filename, 'rb')
        f = BytesIO(fzip.read())
        fzip.close()
    else:
        f = open(filename, 'rb')
        #print(1)

    return f


def _read_header(f):
    """Extract header from a stream and return it
    as a python dict
    """

    prop = {}
    while True:
        key, val = f.readline().decode().rstrip('\n\r').split(" = ")
        if key == 'ElementDataFile':
            if val == 'LOCAL':
                break
            else:
                msg = "unable to read that type of data: '" + str(key) + ' = ' + str(val) + "'"
                raise UserWarning(msg)
        else:
            prop[key] = val

    return prop


def read_metaimage(filename):
    """

    :param filename:
    :return:
    """
    f = open_metaimagefile(filename)

    # read header
    prop = _read_header(f)

    #
    # Jonathan: 14.05.2012
    #
    prop["filename"] = filename

    #
    # find dimensions
    #
    dim = prop.pop("DimSize").split(' ')
    if len(dim) == 2:
        xdim = int(dim[0])
        ydim = int(dim[1])
        zdim = 1
    elif len(dim) == 3:
        xdim = int(dim[0])
        ydim = int(dim[1])
        zdim = int(dim[2])
    else:
        msg = "unable to handle such dimensions: 'DimSize = " + str(len(dim)) + "'"
        raise UserWarning(msg)

    vdim = int(prop.pop("ElementNumberOfChannels", 1))

    #
    # find type
    #
    voxeltype = prop.pop("ElementType")
    if voxeltype == 'MET_UCHAR':
        ntyp = np.dtype(np.uint8)
    elif voxeltype == 'MET_USHORT':
        ntyp = np.dtype(np.uint16)
    elif voxeltype == 'MET_UINT':
        ntyp = np.dtype(np.uint32)
    elif voxeltype == 'MET_FLOAT':
        ntyp = np.dtype(np.float32)
    elif voxeltype == 'MET_DOUBLE':
        ntyp = np.dtype(np.float64)
    else:
        msg = "unable to handle such voxel type: 'ElementType = " + str(voxeltype) + "'"
        raise UserWarning(msg)

    #
    # find resolution
    #
    resolution = prop.pop("ElementSize").split(' ')
    res = [];
    for i in range(0,len(resolution)):
        res.append(float(resolution[i]))

    # read datas
    size = ntyp.itemsize * xdim * ydim * zdim * vdim
    mat = np.fromstring(f.read(size), ntyp)
    if vdim != 1:
        mat = mat.reshape((vdim, xdim, ydim, zdim), order="F")
        mat = mat.transpose(1, 2, 3, 0)
    else:
        mat = mat.reshape((xdim, ydim, zdim), order="F")
        # mat = mat.transpose(2,1,0)

    # create SpatialImage

    img = SpatialImage(mat, res, vdim, prop)

    # return
    f.close()
    return img


def write_metaimage_to_stream(stream, img):
    assert img.ndim in (3, 4)

    # metadata
    info = dict(getattr(img, "info", {}))

    #
    #
    #
    info["ObjectType"] = "Image"

    #
    # image dimensions
    # won't support 2D vectorial images
    #
    if img.ndim == 2:
        info["NDims"] = "3"
        info["DimSize"] = str(img.shape[0]) + ' ' + str(img.shape[1]) + ' 1'
        info["ElementNumberOfChannels"] = "1"
    elif img.ndim == 3:
        info["NDims"] = "3"
        info["DimSize"] = str(img.shape[0]) + ' ' + str(img.shape[1]) + ' ' + str(img.shape[2])
        info["ElementNumberOfChannels"] = "1"
    elif img.ndim == 4:
        info["NDims"] = "3"
        info["DimSize"] = str(img.shape[0]) + ' ' + str(img.shape[1]) + ' ' + str(img.shape[2])
        info["ElementNumberOfChannels"] = str(img.shape[2])

    #
    # image resolutions
    #
    res = getattr(img, "resolution", (1, 1, 1))
    info["ElementSize"] = str(res[0]) + ' ' + str(res[1]) + ' ' + str(res[2])
    info["ElementSpacing"] = str(res[0]) + ' ' + str(res[1]) + ' ' + str(res[2])

    #
    # data type
    #
    if img.dtype == np.uint8:
        info["ElementType"] = "MET_UCHAR"
    elif img.dtype == np.uint16:
        info["ElementType"] = "MET_USHORT"
    elif img.dtype == np.uint32:
        info["ElementType"] = "MET_UINT"
    elif img.dtype == np.float32:
        info["ElementType"] = "MET_FLOAT"
    elif img.dtype == np.float64:
        info["ElementType"] = "MET_DOUBLE"
    # elif img.dtype == np.float128:
    #   info["TYPE"] = "float"
    #   info["PIXSIZE"] = "128 bits"
    else:
        msg = "unable to write that type of data: %s" % str(img.dtype)
        raise UserWarning(msg)

    #
    #
    #
    info["CompressedData"] = "False"
    info["BinaryDataByteOrderMSB"] = "False"
    info["BinaryData"] = "True"
    info["ElementDataFile"] = "LOCAL"

    #
    # fill header
    #
    header = ''
    for k in specific_header_keys:
        try:
            header += "%s = %s\n" % (k, info[k])
        except KeyError:
            pass

    #
    # write raw data
    #

    stream.write(header.encode())
    if img.ndim == 2 or img.ndim == 3:
        stream.write(img.tostring("F"))
    # elif img.ndim == 4:
    #    mat = img.transpose(3,0,1,2)
    #    stream.write(mat.tostring("F") )
    else:
        raise Exception("Unhandled image dimension %d." % img.ndim)


def write_metaimage(filename, img):
    """Write an inrimage zipped or not according to the extension

    .. warning:: if img is not a |SpatialImage|, default values will be used
                 for the resolution of the image

   :Parameters:
     - `img` (|SpatialImage|) - image to write
     - `filename` (str) - name of the file to read
    """
    # open stream
    zipped = (path.splitext(filename)[1] in (".gz", ".zip"))

    if zipped:
        f = gzip.GzipFile(filename, "wb")
        # f = StringIO()
    else:
        f = open(filename, 'wb')

    try:
        write_metaimage_to_stream(f, img)
    except:
        # -- remove probably corrupt file--
        f.close()
        if path.exists(filename) and path.isfile(filename):
            os.remove(filename)
        raise
    else:
        f.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
 
from setuptools import setup, find_packages

import morphonet
 
setup(
 
    name='morphonet',
 
    version="2.0.0.82",
 
    packages=find_packages(),
 
    author="Emmanuel Faure",
 
    author_email="api@morphonet.org",
 
    description="Python API to interact with MorphoNet",
 
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',

    install_requires= [ "requests", "numpy","vtk","imageio","scikit-image","scipy","vtk","nibabel"] ,

    include_package_data=True, # MANIFEST.in
 
    url='https://gitlab.inria.fr/efaure/MorphoNet',
 
    # MetaData
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers.
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Visualization",
    ],

 
)